import { registerAction } from "./history_manager.js";
import { exportGroup, exportInputConnection, exportNode, exportOutputConnection } from "./export.js";
import { importConnections, importNode } from "./import.js";
import { getRelatedConnections } from "./related_connections.js";
import { getProject } from "./repl.js";
import { Normalizer } from "./id_generator.js";
export var Errors;
(function (Errors) {
    Errors[Errors["Undetermined"] = -1] = "Undetermined";
    Errors[Errors["Success"] = 0] = "Success";
    Errors[Errors["NodeNotFound"] = 1] = "NodeNotFound";
    Errors[Errors["SecondNodeNotFound"] = 2] = "SecondNodeNotFound";
    Errors[Errors["PropertyNotFound"] = 3] = "PropertyNotFound";
    Errors[Errors["PortKeyNotFound"] = 4] = "PortKeyNotFound";
    Errors[Errors["SecondPortKeyNotFound"] = 5] = "SecondPortKeyNotFound";
    Errors[Errors["InvalidComponentKey"] = 6] = "InvalidComponentKey";
    Errors[Errors["ModuleNotFound"] = 7] = "ModuleNotFound";
    Errors[Errors["InvalidConnectionId"] = 8] = "InvalidConnectionId";
    Errors[Errors["GroupNotFound"] = 9] = "GroupNotFound";
})(Errors || (Errors = {}));
export var Actions;
(function (Actions) {
    Actions["CreateNode"] = "create-node";
    Actions["RemoveNode"] = "remove-node";
    Actions["CreateConnection"] = "create-connection";
    Actions["RemoveConnection"] = "remove-connection";
    Actions["SetProperty"] = "set-property";
    Actions["RemoveProperty"] = "remove-property";
    Actions["SetMetadata"] = "set-metadata";
    Actions["RemoveMetadata"] = "remove-metadata";
    Actions["MoveNodesToModule"] = "move-nodes-to-module";
    Actions["SwapConnections"] = "swap-connections";
    Actions["CreateGroup"] = "create-group";
    Actions["RemoveGroup"] = "remove-group";
    Actions["AddNodesToGroup"] = "add-nodes-to-group";
    Actions["RemoveNodesFromGroup"] = "remove-nodes-from-group";
    Actions["CopyNodes"] = "copy-nodes";
    Actions["PasteNodes"] = "paste-nodes";
})(Actions || (Actions = {}));
export const ORIGIN_PATCH_KEY = "_origin_patch_";
export const ORIGIN_PROJECT_KEY = "_origin_project_";
let copiedNodeSchemas = [];
export const hasCopiedNodes = () => { return copiedNodeSchemas.length > 0; };
registerAction(Actions.CreateNode, {
    description: "Create Node",
    do: (data, redo) => {
        let node;
        let patch = data.get(ORIGIN_PATCH_KEY);
        if (redo) {
            let id = data.get("id");
            node = patch.createNode(id, data.arg.componentKey, data.arg.properties, data.arg.metadata);
        }
        else {
            let id = data.get(ORIGIN_PROJECT_KEY).idGenerator.next();
            node = patch.createNode(id, data.arg.componentKey, data.arg.properties, data.arg.metadata);
            if (node == null) {
                return Errors.InvalidComponentKey;
            }
            data.set("id", id);
        }
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let id = data.get("id");
        patch.removeNode(id);
    }
});
registerAction(Actions.RemoveNode, {
    description: "Remove Node",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        if (node == null) {
            return Errors.NodeNotFound;
        }
        data.set("nodeSchema", exportNode(node));
        patch.removeNode(data.arg.nodeId);
        return Errors.Success;
    },
    undo: (data) => {
        let nodeSchema = data.get("nodeSchema");
        importNode(nodeSchema, data.get(ORIGIN_PATCH_KEY));
        importConnections(nodeSchema, data.get(ORIGIN_PATCH_KEY));
    }
});
registerAction(Actions.CreateConnection, {
    description: "Create Connection",
    do: (data, redo) => {
        let id;
        let patch = data.get(ORIGIN_PATCH_KEY);
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let toNode = patch.getNode(data.arg.toNodeId);
        let fromPortKey = data.arg.fromPortKey;
        let toPortKey = data.arg.toPortKey;
        let toIndex = data.arg.toIndex;
        if (fromNode == null) {
            return Errors.NodeNotFound;
        }
        if (toNode == null) {
            return Errors.SecondNodeNotFound;
        }
        if (!fromNode.isConnectableOutputPort(fromPortKey)) {
            return Errors.PortKeyNotFound;
        }
        if (!toNode.isConnectableInputPort(toPortKey)) {
            return Errors.SecondPortKeyNotFound;
        }
        if (redo) {
            id = data.get("connectionId");
        }
        else {
            id = data.get(ORIGIN_PROJECT_KEY).idGenerator.next();
            data.set("connectionId", id);
        }
        if (toIndex === undefined) {
            toIndex = toNode.getInputPort(toPortKey).connections.length;
        }
        fromNode.createConnection(id, fromPortKey, toNode, toPortKey, toIndex);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let id = data.get("connectionId");
        fromNode.removeOutputConnection(data.arg.fromPortKey, id);
    }
});
registerAction(Actions.RemoveConnection, {
    description: "Remove Connection",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let fromPort = fromNode.getOutputPort(data.arg.fromPortKey);
        if (!redo) {
            let connection = fromPort.getConnection(data.arg.connectionId);
            data.set("portConnectionsSchema", exportOutputConnection(connection));
            data.set("inputIndex", fromPort.getConnection(data.arg.connectionId).getInputIndex());
        }
        fromNode.removeOutputConnection(data.arg.fromPortKey, data.arg.connectionId);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let schema = data.get("portConnectionsSchema");
        let toNode = patch.getNode(schema.toNodeId);
        let inputIndex = data.get("inputIndex");
        fromNode.createConnection(schema.id, data.arg.fromPortKey, toNode, schema.toPortKey, inputIndex);
    }
});
registerAction(Actions.SetProperty, {
    description: "Set Property",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let propertyKey = data.arg.propertyKey;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!node.component.hasProperty(propertyKey)) {
            return Errors.PropertyNotFound;
        }
        let oldValue = node.isPropertySet(propertyKey) ? node.getProperty(propertyKey) : undefined;
        data.set("oldValue", oldValue);
        node.setProperty(propertyKey, data.arg.value);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        node.removeProperty(data.arg.propertyKey);
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setProperty(data.arg.propertyKey, oldValue);
        }
        return Errors.Success;
    }
});
registerAction(Actions.RemoveProperty, {
    description: "Remove Property",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let propertyKey = data.arg.propertyKey;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!node.component.hasProperty(propertyKey)) {
            return Errors.PropertyNotFound;
        }
        if (!redo) {
            let oldValue = node.isPropertySet(propertyKey) ? node.getProperty(propertyKey) : undefined;
            data.set("oldValue", oldValue);
        }
        node.removeProperty(data.arg.propertyKey);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setProperty(data.arg.propertyKey, oldValue);
        }
        return Errors.Success;
    }
});
registerAction(Actions.SetMetadata, {
    description: "Set Metadata",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let key = data.arg.key;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        let oldValue = node.isMetadataSet(key) ? node.getMetadata(key) : undefined;
        data.set("oldValue", oldValue);
        node.setMetadata(key, data.arg.value);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        //node.removeMetadata(data.arg.key);
        let oldValue = data.get("oldValue");
        //if (oldValue != undefined) {
        node.setMetadata(data.arg.key, oldValue);
        //}
        return Errors.Success;
    }
});
registerAction(Actions.RemoveMetadata, {
    description: "Remove Metadata",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let key = data.arg.key;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!redo) {
            let oldValue = node.isMetadataSet(key) ? node.getMetadata(key) : undefined;
            data.set("oldValue", oldValue);
        }
        node.removeMetadata(data.arg.key);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setMetadata(data.arg.key, oldValue);
        }
        return Errors.Success;
    }
});
registerAction(Actions.MoveNodesToModule, {
    description: "Move Nodes To Module",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let nodes = [];
        let moduleNode = patch.getNode(data.arg.moduleId);
        if (moduleNode == null) {
            return Errors.ModuleNotFound;
        }
        // Collect node references
        for (let id of data.arg.nodeIdList) {
            let node = patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            nodes.push(node);
        }
        let externalConnections = getRelatedConnections(nodes, "external");
        // Remove external connections
        for (let [nodeId, connection] of Object.entries(externalConnections)) {
            for (let conn of connection.inputs) {
                patch.getNode(+nodeId).removeInputConnection(conn.toPortKey, conn.id);
            }
            for (let conn of connection.outputs) {
                patch.getNode(+nodeId).removeOutputConnection(conn.fromPortKey, conn.id);
            }
        }
        data.set("externalConnections", externalConnections);
        // Move nodes to module patch
        for (let node of nodes) {
            moduleNode.patch.nodes[node.id] = node;
            //delete patch.nodes[node.id];
            patch.removeNode(node.id, false);
        }
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let moduleNode = patch.getNode(data.arg.moduleId);
        let nodes = [];
        // Collect node references
        for (let id of data.arg.nodeIdList) {
            let node = moduleNode.patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            nodes.push(node);
        }
        // Move nodes from module back to patch
        for (let node of nodes) {
            patch.nodes[node.id] = node;
            delete moduleNode.patch.nodes[node.id];
        }
        // Restore connections
        let externalConnections = data.get("externalConnections");
        for (let [nodeId, connection] of Object.entries(externalConnections)) {
            for (let conn of connection.inputs) {
                let fromNode = patch.getNode(conn.fromNodeId);
                let toNode = patch.getNode(conn.toNodeId);
                fromNode.createConnection(conn.id, conn.fromPortKey, toNode, conn.toPortKey);
            }
            for (let conn of connection.outputs) {
                let fromNode = patch.getNode(conn.fromNodeId);
                let toNode = patch.getNode(conn.toNodeId);
                fromNode.createConnection(conn.id, conn.fromPortKey, toNode, conn.toPortKey);
            }
        }
        return Errors.Success;
    }
});
registerAction(Actions.SwapConnections, {
    description: "Swap Connections",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        if (node == null) {
            return Errors.NodeNotFound;
        }
        let portKey = data.arg.portKey;
        let index1, index2;
        if (redo) {
            [index1, index2] = data.get("indeces");
        }
        else {
            let port = node.getInputPort(portKey);
            if (port == null || !node.component.hasInputPort(portKey)) {
                return Errors.PortKeyNotFound;
            }
            index1 = port.getConnectionIndex(data.arg.sourceConnectionId);
            index2 = port.getConnectionIndex(data.arg.targetConnectionId);
            if (index1 < 0 || index2 < 0) {
                return Errors.InvalidConnectionId;
            }
        }
        data.set("indeces", [index1, index2]);
        node.swapInputConnections(data.arg.portKey, index1, index2);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let node = patch.getNode(data.arg.nodeId);
        let [index1, index2] = data.get("indeces");
        node.swapInputConnections(data.arg.portKey, index1, index2);
    }
});
registerAction(Actions.CreateGroup, {
    description: "Create Group",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let id;
        if (redo) {
            id = data.get("id");
        }
        else {
            id = getProject().idGenerator.next();
        }
        patch.createGroup(id, data.arg.name || "New Group", "", data.arg.nodeIdList);
        data.set("id", id);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let id = data.get("id");
        patch.removeGroup(id);
    }
});
registerAction(Actions.RemoveGroup, {
    description: "Remove Group",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let id = data.arg.groupId;
        let group = patch.groups[id];
        if (group == null) {
            return Errors.GroupNotFound;
        }
        data.set("groupSchema", exportGroup(group));
        data.set("id", group.id);
        patch.removeGroup(id);
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let groupSchema = data.get("groupSchema");
        let id = data.get("id");
        patch.createGroup(id, groupSchema.name, groupSchema.comment, groupSchema.nodeIdList);
    }
});
registerAction(Actions.AddNodesToGroup, {
    description: "Add Nodes to Group",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let group = patch.groups[data.arg.groupId];
        if (group == null) {
            return Errors.GroupNotFound;
        }
        for (let nodeId of data.arg.nodeIdList) {
            group.addNodeId(nodeId);
        }
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.removeNodeId(nodeId);
        }
    }
});
registerAction(Actions.RemoveNodesFromGroup, {
    description: "Remove Nodes From Group",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.removeNodeId(nodeId);
        }
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.addNodeId(nodeId);
        }
    }
});
registerAction(Actions.CopyNodes, {
    description: "Copy node(s)",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let nodeIdList = data.arg.nodeIdList;
        let nodeSchemas = [];
        for (let id of nodeIdList) {
            let node = patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            let nodeSchema = exportNode(node, { excludeInputs: true, excludeOutputs: true });
            for (let dir of ["in", "out"]) {
                let ports = dir == "in" ? node.inputs : node.outputs;
                let connectionMap = {};
                for (let [portKey, port] of Object.entries(ports)) {
                    let connections = [];
                    for (let [_, connection] of port.connections) {
                        let id = dir == "in"
                            ? connection.outputPort.ownerNode.id
                            : connection.inputPort.ownerNode.id;
                        if (nodeIdList.indexOf(id) >= 0) {
                            connections.push(dir == "in" ? exportInputConnection(connection) : exportInputConnection(connection));
                        }
                    }
                    connectionMap[portKey] = connections;
                }
                if (dir == "in") {
                    nodeSchema.inputs = connectionMap;
                }
                else {
                    nodeSchema.outputs = connectionMap;
                }
            }
            nodeSchemas.push(nodeSchema);
        }
        copiedNodeSchemas = nodeSchemas;
        return Errors.Undetermined; // Not undo/redo-able
    },
    undo: (_) => {
    }
});
registerAction(Actions.PasteNodes, {
    description: "Paste nodes(s)",
    do: (data, redo) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        let nodeSchemas;
        let normalizer = new Normalizer(getProject().idGenerator);
        if (redo) {
            nodeSchemas = data.get("copiedNodeSchemas");
        }
        else {
            nodeSchemas = copiedNodeSchemas;
            data.set("copiedNodeSchemas", copiedNodeSchemas);
        }
        let createdNodeIdList = [];
        for (let nodeSchema of copiedNodeSchemas) {
            let node = importNode(nodeSchema, patch, normalizer);
            createdNodeIdList.push(node.id);
        }
        data.set("createdNodeIdList", createdNodeIdList);
        for (let nodeSchema of nodeSchemas) {
            importConnections(nodeSchema, patch, normalizer);
        }
        return Errors.Success;
    },
    undo: (data) => {
        let patch = data.get(ORIGIN_PATCH_KEY);
        for (let id of data.consume("createdNodeIdList")) {
            patch.removeNode(id);
        }
    }
});
//# sourceMappingURL=actions.js.map