import { Errors } from "./actions.js";
import { Events } from "./events.js";
import EventHandler from "./event_handler.js";
import observer from "./observer.js";
const MAX_HISTORY_LENGTH = 50;
export class ActionData {
    constructor(arg) {
        this.data = {};
        this.arg = arg;
    }
    set(name, value) {
        this.data[name] = value;
    }
    get(name) {
        return this.data[name];
    }
    consume(name) {
        let value = this.get(name);
        delete this.data[name];
        return value;
    }
}
const actions = {};
export const registerAction = (key, action) => {
    actions[key] = action;
};
export class HistoryManager extends EventHandler {
    constructor() {
        super(...arguments);
        this.entries = [];
        this.historyIndex = 0;
        this.batch = [];
    }
    redo() {
        let result = false;
        if (this.canRedo()) {
            let entry = this.entries[this.historyIndex++];
            if (entry.entries) {
                // Action batch.
                for (let batchEntry of entry.entries) {
                    let action = actions[batchEntry.key];
                    this.triggerEvent("onBeforeDo", batchEntry.data);
                    result = action.do(batchEntry.data, true);
                    observer.triggerEvent(Events.OnDoAction, entry.data);
                }
                observer.triggerEvent(Events.OnFinishBatchAction, entry);
            }
            else {
                let action = actions[entry.key];
                this.triggerEvent("onBeforeDo", entry.data);
                result = action.do(entry.data, true);
                observer.triggerEvent(Events.OnDoAction, entry.data);
            }
            observer.triggerEvent(Events.OnUpdateData);
        }
        return result;
    }
    undo() {
        let result = false;
        if (this.canUndo()) {
            let history = this.entries[--this.historyIndex];
            if (history.entries) {
                let batchHistory = history.entries;
                for (let i = batchHistory.length - 1; i >= 0; i--) {
                    let entry = batchHistory[i];
                    let action = actions[entry.key];
                    this.triggerEvent("onBeforeUndo", entry.data);
                    result = action.undo(entry.data);
                    observer.triggerEvent(Events.OnUndoAction, history.data);
                }
                observer.triggerEvent(Events.OnUndoBatchAction, history);
            }
            else {
                let action = actions[history.key];
                this.triggerEvent("onBeforeUndo", history.data);
                result = action.undo(history.data);
                observer.triggerEvent(Events.OnUndoAction, history.data);
            }
            observer.triggerEvent(Events.OnUpdateData);
        }
        return result;
    }
    pushAction(name, data) {
        this.entries.length = this.historyIndex++;
        this.entries.push({
            key: name,
            data: data
        });
        if (this.entries.length > MAX_HISTORY_LENGTH) {
            this.entries.shift();
        }
    }
    doAction(name, arg, callback) {
        let action = actions[name];
        //let data = new ActionData(arg);
        let data = new ActionData(arg);
        this.triggerEvent("onCreateActionData", data);
        this.triggerEvent("onBeforeDo", data);
        let result = action.do(data);
        if (result == Errors.Success) {
            this.pushAction(name, data);
            observer.triggerEvent(Events.OnDoAction, data);
            observer.triggerEvent(Events.OnUpdateData);
        }
        callback && callback(data);
        return result;
    }
    doBatchAction(key, arg, callback) {
        let action = actions[key];
        let data = new ActionData(arg);
        this.triggerEvent("onCreateActionData", data);
        this.triggerEvent("onBeforeDo", data);
        let result = action.do(data);
        this.batch.push({
            key: key,
            data: data
        });
        observer.triggerEvent(Events.OnDoAction, data);
        observer.triggerEvent(Events.OnUpdateData);
        callback && callback(data);
        return result;
    }
    finishBatch(key, description = "") {
        if (this.batch.length > 0) {
            this.entries.length = this.historyIndex++;
            const entry = {
                "description": description,
                key: key,
                entries: [...this.batch]
            };
            this.entries.push(entry);
            observer.triggerEvent(Events.OnFinishBatchAction, entry);
            this.batch.length = 0;
            if (this.entries.length > MAX_HISTORY_LENGTH) {
                this.entries.shift();
            }
            observer.triggerEvent(Events.OnUpdateData);
        }
    }
    resetBatch() {
        this.batch.length = 0;
    }
    cancelBatch() {
        this.batch = [];
    }
    canUndo() {
        return this.historyIndex > 0;
    }
    canRedo() {
        return this.entries.length > this.historyIndex;
    }
    reset() {
        this.entries.length = 0;
        this.historyIndex = 0;
        observer.triggerEvent(Events.OnUpdateData);
    }
    getAction(name) {
        return actions[name];
    }
}
//# sourceMappingURL=history_manager.js.map