import observer from "./observer.js";
import Port, { PortType } from "./port.js";
import { getComponent } from "./component_manager.js";
import { FLOW_PORT_KEY } from "./component.js";
import { Events } from "./events.js";
export var NodeType;
(function (NodeType) {
    NodeType[NodeType["Normal"] = 0] = "Normal";
    NodeType[NodeType["Module"] = 1] = "Module";
    NodeType[NodeType["Group"] = 2] = "Group";
    NodeType[NodeType["ModuleInput"] = 3] = "ModuleInput";
    NodeType[NodeType["ModuleOutput"] = 4] = "ModuleOutput";
    NodeType[NodeType["Internal"] = 5] = "Internal";
})(NodeType || (NodeType = {}));
class Node {
    constructor(ownerPatch, id, componentIdentKey) {
        this.type = NodeType.Normal;
        this.name = "";
        this.outputs = {};
        this.inputs = {};
        this.properties = {};
        this.metadata = {};
        this.ownerPatch = ownerPatch;
        this.id = id;
        this.componentIdentKey = componentIdentKey;
        this.component = getComponent(componentIdentKey);
    }
    isConnectableInputPort(key) {
        return this.component.hasInputPort(key);
    }
    isConnectableOutputPort(key) {
        return this.component.hasOutputPort(key);
    }
    createFlowPort() {
        return {
            name: "Flow",
            type: "flow"
        };
    }
    getConnectableInputPorts() {
        let flowPort = [FLOW_PORT_KEY, this.createFlowPort()];
        if (this.component.data.inputs != null) {
            let inputs = Object.entries(this.component.data.inputs);
            inputs.push(flowPort);
            let filtered = inputs.filter(([portKey, portProperties]) => {
                if (portProperties.isMulti) {
                    return true;
                }
                return !((portKey in this.inputs) && this.inputs[portKey].connections.length > 0);
            });
            return filtered.map(([key, data]) => {
                return [key, data.name];
            });
        }
        return [flowPort];
    }
    getConnectableOutputPorts() {
        let flowPort = [FLOW_PORT_KEY, this.createFlowPort()];
        if (this.component.data.outputs != null) {
            let outputs = Object.entries(this.component.data.outputs);
            outputs.push(flowPort);
            let filtered = outputs.filter(([key, data]) => {
                return !((key in this.outputs) && this.outputs[key].connections.length > 0);
            });
            return filtered.map(([key, data]) => {
                return [key, data.name];
            });
        }
        return [flowPort];
    }
    lazyGetPort(type, portKey) {
        let portMap = type == PortType.Output ? this.outputs : this.inputs;
        if (portMap[portKey] == null) {
            return portMap[portKey] = new Port(portKey, type, this);
        }
        return portMap[portKey];
    }
    getOutputPort(portKey) {
        return this.lazyGetPort(PortType.Output, portKey);
    }
    getInputPort(portKey) {
        var _a;
        let type = ((_a = this.component.getInputPort(portKey)) === null || _a === void 0 ? void 0 : _a.isMulti) ? PortType.InputMulti : PortType.Input;
        return this.lazyGetPort(type, portKey);
    }
    findConnection(connectionId) {
        for (let ports of [this.inputs, this.outputs]) {
            for (let port of Object.values(ports)) {
                let connection = port.getConnection(connectionId);
                if (connection != null) {
                    return connection;
                }
            }
        }
        return null;
    }
    removeOldInputConnection(targetNode, targetPortKey) {
        let targetPort = targetNode.getInputPort(targetPortKey);
        if (targetPort.type == PortType.Input && targetPort.hasConnections()) {
            targetPort.ownerNode.removeInputConnection(targetPortKey, targetPort.connections[0][0]);
        }
    }
    createConnection(id, fromPortKey, targetNode, targetPortKey, inputIndex) {
        this.removeOldInputConnection(targetNode, targetPortKey);
        let connection = this.getOutputPort(fromPortKey).createConnection(id, targetNode.getInputPort(targetPortKey), inputIndex);
        observer.triggerEvent(Events.OnCreateConnection, {
            connection: connection,
            inputIndex: inputIndex
        });
        return connection;
    }
    removeOutputConnection(portKey, connectionId) {
        let removedConnections = [];
        if (connectionId == null) {
            for (let [connectionId, connection] of this.outputs[portKey].connections) {
                connection.inputPort.removeConnection(+connectionId);
                this.outputs[portKey].removeConnection(+connectionId);
                removedConnections.push(connection);
                observer.triggerEvent(Events.OnRemoveOutputConnection, connection);
            }
            ;
        }
        else {
            let connection = this.outputs[portKey].getConnection(connectionId);
            connection.inputPort.removeConnection(connectionId);
            this.outputs[portKey].removeConnection(connectionId);
            removedConnections.push(connection);
            observer.triggerEvent(Events.OnRemoveOutputConnection, connection);
        }
        return removedConnections;
    }
    removeInputConnection(portKey, connectionId) {
        if (connectionId == null) {
            for (let [connectionId, connection] of this.inputs[portKey].connections) {
                connection.outputPort.removeConnection(+connectionId);
                this.inputs[portKey].removeConnection(+connectionId);
                observer.triggerEvent(Events.OnRemoveInputConnection, connection);
            }
            ;
        }
        else {
            let connection = this.inputs[portKey].getConnection(connectionId);
            connection.outputPort.removeConnection(connectionId);
            this.inputs[portKey].removeConnection(connectionId);
            observer.triggerEvent(Events.OnRemoveInputConnection, connection);
        }
    }
    swapInputConnections(portKey, index1, index2) {
        let port = this.inputs[portKey];
        let conns = port.connections;
        [conns[index1], conns[index2]] = [conns[index2], conns[index1]];
        observer.triggerEvent(Events.OnSwapInputConnections, {
            port: port,
            index1: index1,
            index2: index2
        });
    }
    isPropertySet(key) {
        return key in this.properties;
    }
    setProperty(key, value) {
        this.properties[key] = value;
        observer.triggerEvent(Events.OnSetProperty, {
            node: this,
            key: key,
            value: value
        });
    }
    getProperty(key) {
        var _a, _b;
        let value;
        if (this.isPropertySet(key)) {
            value = this.properties[key];
        }
        else {
            let component = getComponent(this.componentIdentKey);
            value = (_b = (_a = component === null || component === void 0 ? void 0 : component.data.properties) === null || _a === void 0 ? void 0 : _a[key]) === null || _b === void 0 ? void 0 : _b.defaultValue;
        }
        return value;
    }
    removeProperty(key) {
        let value = this.properties[key];
        delete this.properties[key];
        observer.triggerEvent(Events.OnRemoveProperty, {
            node: this,
            key: key,
            value: value
        });
    }
    isMetadataSet(key) {
        return key in this.metadata;
    }
    setMetadata(key, value) {
        if (typeof value === "undefined") {
            this.removeMetadata(key);
        }
        else {
            this.metadata[key] = value;
        }
        observer.triggerEvent(Events.OnSetMetadata, {
            node: this,
            key: key,
            value: value
        });
    }
    getMetadata(key) {
        return this.metadata[key];
    }
    removeMetadata(key) {
        let value = this.metadata[key];
        delete this.metadata[key];
        observer.triggerEvent(Events.OnRemoveMetadata, {
            node: this,
            key: key,
            value: value
        });
    }
    getName() {
        return this.name || this.component.getName();
    }
}
export default Node;
//# sourceMappingURL=node.js.map