import { getNamespaceTree, namespaces } from "./component_manager.js";
import { FLOW_PORT_KEY } from "./component.js";
import { Errors } from "./actions.js";
import { NodeType } from "./node.js";
import Project from "./project.js";
const CORE_MODULE_KEY = "core.module";
var TokenType;
(function (TokenType) {
    TokenType[TokenType["String"] = 0] = "String";
    TokenType[TokenType["Object"] = 1] = "Object";
    TokenType[TokenType["Array"] = 2] = "Array";
    TokenType[TokenType["Number"] = 3] = "Number";
    TokenType[TokenType["Keyword"] = 4] = "Keyword";
})(TokenType || (TokenType = {}));
export const commands = [];
let project = new Project("New project");
export const getProject = () => {
    return project;
};
export const setProject = (setProject) => {
    project = setProject;
};
export let lastErrorMessage = "";
export let print = (text) => {
    console.log(text);
};
export const setPrintFn = (fn) => {
    print = fn;
};
export const fail = (msg) => {
    lastErrorMessage = msg;
    return true;
};
const convertFlowPortKey = (key) => {
    return key == "flow" ? FLOW_PORT_KEY : key == FLOW_PORT_KEY ? "flow" : key;
};
export const parseCmdLine = (cmd) => {
    let resultTokens = [];
    let token = "";
    let closureParse = false;
    let closureEndChars = [];
    for (let char of cmd) {
        if (token == "" && (char == " " || char == "\n") && !closureParse) {
            continue;
        }
        else if (!closureParse && (char == "\"" || char == "[" || char == "{")) {
            closureParse = true;
            closureEndChars.push(char == "\"" ? "\"" : char == "[" ? "]" : "}");
            if (char == "\"") {
                continue;
            }
        }
        else if (closureParse) {
            if (char == closureEndChars[closureEndChars.length - 1]) {
                closureEndChars.pop();
                if (closureEndChars.length == 0) {
                    resultTokens.push({
                        text: token + (char != "\"" ? char : ""),
                        type: char == "\"" ? TokenType.String : char == "]" ? TokenType.Array : TokenType.Object
                    });
                    token = "";
                    closureParse = false;
                }
                continue;
            }
        }
        else if ((char == " " || char == "\n") && !closureParse) {
            resultTokens.push({
                text: token,
                type: isNumber(token) ? TokenType.Number : TokenType.Keyword
            });
            token = "";
            continue;
        }
        token += char;
    }
    if (token.length > 0) {
        resultTokens.push({
            text: token,
            type: isNumber(token) ? TokenType.Number : TokenType.Keyword
        });
    }
    return resultTokens;
};
let pathHistory = [];
const updatePathHistory = () => {
    let ownerNode = project.getActivePatch().ownerModuleNode;
    if (ownerNode == null) {
        return;
    }
    let parentPatch = ownerNode.ownerPatch;
    let len = pathHistory.length;
    for (let i = 0; i < len; i++) {
        let patch = pathHistory[i];
        if (patch.id == project.getActivePatch().id) {
            return;
        }
        else if (patch.id == parentPatch.id && i < len - 1) {
            if (pathHistory[i + 1].id != project.getActivePatch().id) {
                pathHistory.splice(i + 1);
                break;
            }
        }
    }
    pathHistory.push(project.getActivePatch());
};
const nextPatchInPath = () => {
    let len = pathHistory.length;
    if (!project.getActivePatch().isModule() && len > 0) {
        project.setActivePatch(pathHistory[0]);
        return;
    }
    for (let i = 0; i < len; i++) {
        if (pathHistory[i].id == project.getActivePatch().id && i < len - 1) {
            project.setActivePatch(pathHistory[i + 1]);
            break;
        }
    }
};
const getModulePathStr = () => {
    let pathStr = "";
    for (let patch of pathHistory) {
        let node = patch.ownerModuleNode;
        pathStr += " > ";
        if (node.patch.id == project.getActivePatch().id) {
            pathStr += `[${node.name}]`;
        }
        else {
            pathStr += `${node.name}`;
        }
    }
    return (project.getActivePatch().ownerModuleNode == null ? "[Patch]" : "Patch") + pathStr;
};
export const registerCommand = (name, validators, autocomplete, fn) => {
    commands.push({
        name: name,
        validators: validators,
        autocomplete: autocomplete,
        fn: fn
    });
};
export const findCommand = (name) => {
    for (let entry of commands) {
        if (entry.name.indexOf(name) >= 0) {
            return entry;
        }
    }
    return null;
};
export const executeCommand = (cmdLine) => {
    let failed = false;
    //const cmdLineArr = cmdLine.split(" ").map(x => x.trim()).filter(x => x !== "");
    const tokens = parseCmdLine(cmdLine);
    if (tokens.length == 0) {
        return false;
    }
    const cmdName = tokens[0].text;
    let command = findCommand(cmdName);
    if (!command) {
        return fail(`Unknown command. Type 'help' for available commands.`);
    }
    ;
    let validatorsArr = Object.entries(command.validators);
    if (validatorsArr.length == 0) {
        return command.fn(tokens.slice(1));
    }
    let numArgsFailMsg = "";
    for (let [validatorKey, validator] of validatorsArr) {
        let argKeys = Object.keys(validator);
        let numInputArgs = tokens.length - 1;
        if (argKeys.length != numInputArgs) {
            numArgsFailMsg = `Wrong number of arguments (${numInputArgs}) expected (${argKeys.length})`;
            continue;
        }
        let argValidators = command.validators[validatorKey];
        for (let i = 0; i < tokens.length - 1; i++) {
            let token = tokens[i + 1];
            let key = argKeys[i];
            let result = argValidators[key](token);
            if (result !== true) {
                failed = true;
                break;
            }
            else {
                failed = false;
            }
        }
        if (!failed) {
            return command.fn(tokens.slice(1), validatorKey);
        }
    }
    if (!failed && numArgsFailMsg != "") {
        return fail(numArgsFailMsg);
    }
    return false;
};
const validateNodeName = (s) => {
    return (/^[a-z0-9]+$/i).test(s);
};
const validatePortKey = (s) => {
    return (/^[a-z0-9]+$/i).test(s);
};
const validateComponentKey = (s) => {
    let e = s.split(".");
    for (let x of e) {
        if (!(/^[a-z0-9]+$/i).test(x))
            return false;
    }
    return true;
};
const isNumber = (s) => {
    return (/^-?\d+\.?\d*$/).test(s);
};
const validateValue = (s) => {
    return true;
};
/*
 Commands
 */
// List
registerCommand(
// Name
["list", "ls"], 
// Validators
{
    "no-params": {},
    "list-mode": {
        0: (token) => {
            return ((token.type != TokenType.Keyword && token.type != TokenType.Number) ||
                (!isNumber(token.text) && (["all", "inputs", "outputs", "path"]).indexOf(token.text) < 0))
                ? fail(`Unexpected argument '${token.text}'`)
                : true;
        },
    },
}, 
// Autocomplete
["all", "inputs", "outputs", "path"], 
// Command fn
(tokens) => {
    let patch = project.getActivePatch();
    let mode = "outputs";
    if (tokens.length == 0) {
    }
    else {
        let token = tokens[0];
        if (token.type == TokenType.Number) {
            let nodeId = parseInt(token.text);
            let node = project.getActivePatch().nodes[nodeId];
            if (node == null) {
                return fail(`There is no node with id '${nodeId}'`);
            }
            if (node.type != 1) {
                return fail(`Node with id '${nodeId}' is not an module`);
            }
            patch = node.ownerPatch;
        }
        else {
            mode = token.text;
        }
    }
    let modulePathStr = getModulePathStr();
    print("-".repeat(Math.max(modulePathStr.length, 32)));
    print(modulePathStr);
    print("-".repeat(Math.max(modulePathStr.length, 32)));
    if (mode == "path") {
        return false;
    }
    for (let [id, node] of Object.entries(patch.nodes)) {
        // Properties
        let propertyStr = "";
        let component = node.component; //getComponent(node.componentIdentKey);
        if (component.data.properties != null) {
            let props = Object.entries(component.data.properties);
            if (props.length > 0) {
                propertyStr += " : ";
                let count = 0;
                for (let [key, data] of props) {
                    if (count > 0) {
                        propertyStr += ", ";
                    }
                    let val = node.getProperty(key);
                    propertyStr += `${key}=${val}`;
                    count++;
                }
            }
        }
        print(`${id}: [${node.componentIdentKey}]` + propertyStr);
        // Metadata
        if (Object.keys(node.metadata).length > 0) {
            let metadataStr = "";
            for (let [key, value] of Object.entries(node.metadata)) {
                let valueStr = "";
                if (typeof value == "object") {
                    valueStr = JSON.stringify(value);
                }
                else {
                    valueStr = value;
                }
                metadataStr += (metadataStr == "" ? "" : ", ") + `${key}=${valueStr}`;
            }
            print(`    Metadata: ${metadataStr}`);
        }
        // Input ports
        if (mode == "inputs" || mode == "in" || mode == "all") {
            for (let [portKey, port] of Object.entries(node.inputs)) {
                let index = 0;
                for (let [id, connection] of port.connections) {
                    let outputPort = connection.outputPort;
                    let indexStr = port.type == 2 ? `[${index}] ` : "";
                    print(`    ${id}: ${indexStr}${convertFlowPortKey(portKey)} <- ${outputPort.ownerNode.id}.${convertFlowPortKey(outputPort.key)}`);
                    index++;
                }
            }
            // Not connected ports
            let connectableInputs = node.getConnectableInputPorts();
            if (connectableInputs.length > 0) {
                let str = "";
                for (let [portKey, name] of connectableInputs) {
                    str += (str != "" ? ", " : "") + `${convertFlowPortKey(portKey)}`;
                }
                print(`    <- ${str}`);
            }
        }
        // Output ports
        if (mode == "outputs" || mode == "out" || mode == "all") {
            for (let [portKey, port] of Object.entries(node.outputs)) {
                for (let [id, connection] of port.connections) {
                    let inputPort = connection.inputPort;
                    print(`    ${id}: ${convertFlowPortKey(portKey)} -> ${inputPort.ownerNode.id}.${convertFlowPortKey(inputPort.key)}`);
                }
            }
            // Not connected ports
            let connectableOutputs = node.getConnectableOutputPorts();
            if (connectableOutputs.length > 0) {
                let str = "";
                for (let [portKey, name] of connectableOutputs) {
                    str += (str != "" ? ", " : "") + `${convertFlowPortKey(portKey)}`;
                }
                print(`    -> ${str}`);
            }
        }
    }
    return false;
});
// Create node
registerCommand(["create-node", "cn"], {
    "validator1": {
        0: (token) => {
            return !validateComponentKey(token.text) ? fail(`Invalid component key '${token.text}'`) : true;
        },
    },
}, 
// Auto completion
() => {
    let components = [];
    for (let [nsKey, ns] of Object.entries(namespaces)) {
        for (let comKey of Object.keys(ns.components)) {
            components.push(`${nsKey}.${comKey}`);
        }
    }
    return components;
}, (tokens) => {
    let componentKey = tokens[0].text;
    let result = project.historyManager.doAction("create-node", { componentKey: componentKey });
    if (result == Errors.InvalidComponentKey) {
        return fail(`There is no component with the key '${componentKey}'`);
    }
    return false;
});
// Remove node
registerCommand(["remove-node", "rn"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    let nodeId = parseInt(tokens[0].text);
    let result = project.historyManager.doAction("remove-node", { nodeId: nodeId });
    if (result == Errors.NodeNotFound) {
        return fail(`There is no node with id '${nodeId}'`);
    }
    return false;
});
// Create module
registerCommand(["create-module", "cm"], {
    "validator1": {},
    "validator2": {
        0: (token) => {
            return !validateNodeName(token.text) ? fail(`Invalid module name '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    //let name = tokens.length > 0 ? tokens[0].text : "New Module";
    project.historyManager.doAction("create-node", { componentKey: CORE_MODULE_KEY });
    return false;
});
// Select module
registerCommand(["open-module", "om"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid module id '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    let nodeId = parseInt(tokens[0].text);
    let node = project.getActivePatch().getNode(nodeId);
    if (node == null) {
        return fail(`There is no module with id '${nodeId}'`);
    }
    if (node.type != NodeType.Module) {
        return fail(`Node with id '${nodeId}' is not a module`);
    }
    project.setActivePatch(node.patch);
    updatePathHistory();
    return false;
});
// Move to module
registerCommand(["move-to-module", "mm"], {}, null, (tokens) => {
    if (tokens.length < 2) {
        return fail(`Too few arguments`);
    }
    let moduleId = parseInt(tokens[0].text);
    let nodeIdList = [];
    for (let i = 1; i < tokens.length; i++) {
        let nodeId = parseInt(tokens[i].text);
        let node = project.getActivePatch().getNode(nodeId);
        if (node == null) {
            return fail(`There is no node with id '${nodeId}'`);
        }
        nodeIdList.push(node.id);
    }
    project.historyManager.doAction("move-nodes-to-module", {
        moduleId: moduleId,
        nodeIdList: nodeIdList
    });
    return false;
});
// Create group
const createNodeIdList = (tokens) => {
    let nodeIdList = [];
    for (let i = 1; i < tokens.length; i++) {
        let nodeId = parseInt(tokens[i].text);
        let node = project.getActivePatch().getNode(nodeId);
        if (node == null) {
            return fail(`There is no node with id '${nodeId}'`);
        }
        nodeIdList.push(node.id);
    }
    return nodeIdList;
};
registerCommand(["create-group", "cg"], {}, null, (tokens) => {
    if (tokens.length < 1) {
        return fail(`Too few arguments`);
    }
    let groupName = tokens[0].text || "New Group";
    let nodeIdList = createNodeIdList(tokens);
    if (nodeIdList === false) {
        return false;
    }
    // let nodeIdList: number[] = [];
    // for (let i = 1; i < tokens.length; i++) {
    //     let nodeId = parseInt(tokens[i].text);
    //     let node = project.getActivePatch().getNode(nodeId);
    //     if (node == null) {
    //         return fail(`There is no node with id '${nodeId}'`);
    //     }
    //     nodeIdList.push(node.id);
    // }
    project.historyManager.doBatchAction("create-group", { name: groupName }, (data) => {
        if (nodeIdList.length > 0) {
            project.historyManager.doBatchAction("add-nodes-to-group", {
                groupId: data.get("id"),
                nodeIdList: nodeIdList
            });
        }
        project.historyManager.finishBatch("create-group", "Create Group");
    });
    return false;
});
// Remove group
registerCommand(["remove-group", "rg"], {
    "validator1": {
        0: (token) => {
            return TokenType.Number ? fail(`Invalid group id '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    let groupId = parseInt(tokens[0].text);
    let result = project.historyManager.doAction("remove-group", {
        groupId: groupId
    });
    if (result == Errors.GroupNotFound) {
        return fail(`There is no group with id '${groupId}'`);
    }
    return false;
});
// Add nodes to group
registerCommand(["add-nodes-to-group", "atg"], {}, null, (tokens) => {
    if (tokens.length < 1) {
        return fail(`Too few arguments`);
    }
    let groupId = parseInt(tokens[0].text);
    if (project.getActivePatch().groups[groupId] == null) {
        return fail(`There is no group with id '${groupId}'`);
    }
    let nodeIdList = createNodeIdList(tokens);
    if (nodeIdList === false) {
        return false;
    }
    // let nodeIdList: number[] = [];
    // for (let i = 1; i < tokens.length; i++) {
    //     let nodeId = parseInt(tokens[i].text);
    //     let node = project.getActivePatch().getNode(nodeId);
    //     if (node == null) {
    //         return fail(`There is no node with id '${nodeId}'`);
    //     }
    //     nodeIdList.push(node.id);
    // }
    if (nodeIdList.length > 0) {
        project.historyManager.doAction("add-nodes-to-group", {
            groupId: groupId,
            nodeIdList: nodeIdList
        });
    }
    return false;
});
// Remove nodes from group
registerCommand(["remove-nodes-from-group", "rfg"], {}, null, (tokens) => {
    if (tokens.length < 1) {
        return fail(`Too few arguments`);
    }
    let groupId = parseInt(tokens[0].text);
    if (project.getActivePatch().groups[groupId] == null) {
        return fail(`There is no group with id '${groupId}'`);
    }
    let nodeIdList = createNodeIdList(tokens);
    if (nodeIdList === false) {
        return false;
    }
    // let nodeIdList: number[] = [];
    // for (let i = 1; i < tokens.length; i++) {
    //     let nodeId = parseInt(tokens[i].text);
    //     let node = project.getActivePatch().getNode(nodeId);
    //     if (node == null) {
    //         return fail(`There is no node with id '${nodeId}'`);
    //     }
    //     nodeIdList.push(node.id);
    // }
    if (nodeIdList.length > 0) {
        project.historyManager.doAction("remove-nodes-from-group", {
            groupId: groupId,
            nodeIdList: nodeIdList
        });
    }
    return false;
});
// List Groups
registerCommand(["list-groups", "lg"], {}, null, (_) => {
    let patch = project.getActivePatch();
    for (let [groupId, group] of Object.entries(patch.groups)) {
        print(`${groupId}: ${group.name}`);
        for (let nodeId of group.nodeIdList) {
            let node = patch.nodes[nodeId];
            print(`  ${nodeId}, ${node.componentIdentKey}`);
        }
    }
    return false;
});
// Prev
registerCommand(["prev", "p"], {}, null, (_) => {
    var _a;
    let prevPatch = (_a = project.getActivePatch().ownerModuleNode) === null || _a === void 0 ? void 0 : _a.ownerPatch;
    if (prevPatch) {
        project.setActivePatch(prevPatch);
    }
    return false;
});
// Next
registerCommand(["next", "n"], {}, null, (_) => {
    nextPatchInPath();
    return false;
});
// Create connection
registerCommand(["create-connection", "cc"], {
    "validator1": {
        0: (nodeId) => {
            return nodeId.type != TokenType.Number ? fail(`Invalid node id '${nodeId.text}'`) : true;
        },
        1: (portKey) => {
            return !validatePortKey(portKey.text) ? fail(`Invalid output port key '${portKey.text}'`) : true;
        },
        2: (nodeId) => {
            return nodeId.type != TokenType.Number ? fail(`Invalid node id '${nodeId.text}'`) : true;
        },
        3: (portKey) => {
            return !validatePortKey(portKey.text) ? fail(`Invalid input port key '${portKey}'`) : true;
        },
    },
}, null, (tokens) => {
    let fromNodeId = tokens[0].text;
    let fromPortKey = tokens[1].text;
    let toNodeId = tokens[2].text;
    let toPortKey = tokens[3].text;
    let result = project.historyManager.doAction("create-connection", {
        fromNodeId: +fromNodeId,
        fromPortKey: convertFlowPortKey(fromPortKey),
        toNodeId: +toNodeId,
        toPortKey: convertFlowPortKey(toPortKey)
    });
    if (result == Errors.NodeNotFound) {
        return fail(`There is no node with id '${fromNodeId}'`);
    }
    if (result == Errors.SecondNodeNotFound) {
        return fail(`There is no node with id '${toNodeId}'`);
    }
    if (result == Errors.PortKeyNotFound) {
        let fromNode = project.getActivePatch().getNode(parseInt(fromNodeId));
        return fail(`Component "${fromNode.componentIdentKey}" has no output port with the key '${fromPortKey}'`);
    }
    if (result == Errors.SecondPortKeyNotFound) {
        let toNode = project.getActivePatch().getNode(parseInt(toNodeId));
        return fail(`Component "${toNode.componentIdentKey}" has no input port with the key '${toPortKey}'`);
    }
    return false;
});
// Remove connection
registerCommand(["remove-connection", "rc"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid connection id '${token}'`) : true;
        },
    },
}, null, (tokens) => {
    let connectionId = parseInt(tokens[0].text);
    let patch = project.getActivePatch();
    let connection = patch.findConnection(connectionId);
    let result = -1;
    if (connection) {
        result = project.historyManager.doAction("remove-connection", {
            fromNodeId: connection.outputPort.ownerNode.id,
            fromPortKey: connection.outputPort.key,
            connectionId: connection.id
        });
    }
    if (result != Errors.Success) {
        return fail(`There is no connection with id '${connectionId}'`);
    }
    return false;
});
// Set property
registerCommand(["set-property", "sp"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token}'`) : true;
        },
        1: (token) => {
            return !validatePortKey(token.text) ? fail(`Invalid property name '${token}'`) : true;
        },
        2: (token) => {
            return !validateValue(token.text) ? fail(`Invalid value: ${token}`) : true;
        },
    },
}, null, (tokens) => {
    let nodeId = tokens[0].text;
    let propertyKey = tokens[1].text;
    let value = tokens[2].text;
    let result = project.historyManager.doAction("set-property", {
        nodeId: nodeId,
        propertyKey: propertyKey,
        value: value
    });
    if (result != Errors.Success) {
        if (result == Errors.NodeNotFound) {
            return fail(`There is no node with id '${nodeId}'`);
        }
        else if (result === Errors.PropertyNotFound) {
            let node = project.getActivePatch().getNode(parseInt(nodeId));
            return fail(`Component '${node.componentIdentKey}' has no property with the key '${propertyKey}'`);
        }
    }
    return false;
});
// Remove prperty
registerCommand(["remove-property", "rp"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token}'`) : true;
        },
        1: (token) => {
            return !validatePortKey(token.text) ? fail(`Invalid property name '${token}'`) : true;
        },
    },
}, null, (tokens) => {
    let nodeId = parseInt(tokens[0].text);
    let propertyKey = tokens[1].text;
    let result = project.historyManager.doAction("remove-property", {
        nodeId: nodeId,
        propertyKey: propertyKey,
    });
    if (result != Errors.Success) {
        if (result == Errors.NodeNotFound) {
            return fail(`There is no node with id '${nodeId}'`);
        }
        else if (result === Errors.PropertyNotFound) {
            let node = project.getActivePatch().getNode(nodeId);
            return fail(`Component '${node.componentIdentKey}' has no property with the key '${propertyKey}'`);
        }
    }
    return false;
});
// Set metadata
registerCommand(["set-metadata", "sm"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token.text}'`) : true;
        },
        1: (token) => {
            return !validatePortKey(token.text) ? fail(`Invalid metadata name '${token.text}'`) : true;
        },
        2: (token) => {
            return !validateValue(token.text) ? fail(`Invalid value: ${token.text}`) : true;
        }
    },
}, null, (tokens) => {
    let nodeId = parseInt(tokens[0].text);
    let key = tokens[1].text;
    let valueToken = tokens[2];
    let value;
    if (valueToken.type == TokenType.Object || valueToken.type == TokenType.Array) {
        console.log(`"${valueToken.text}"`);
        value = JSON.parse(valueToken.text);
    }
    else if (valueToken.type == TokenType.Number) {
        value = parseInt(valueToken.text);
    }
    else {
        value = valueToken.text;
    }
    let result = project.historyManager.doAction("set-metadata", {
        nodeId: nodeId,
        key: key,
        value: value
    });
    if (result == Errors.NodeNotFound) {
        return fail(`There is no node with id '${nodeId}'`);
    }
    return false;
});
// Remove metadata
registerCommand(["remove-metadata", "rm"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token.text}'`) : true;
        },
        1: (token) => {
            return !validatePortKey(token.text) ? fail(`Invalid metadata name '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    let [nodeId, key] = tokens;
    let result = project.historyManager.doAction("remove-metadata", {
        nodeId: nodeId.text,
        key: key.text,
    });
    if (result == Errors.NodeNotFound) {
        return fail(`There is no node with id '${nodeId.text}'`);
    }
    return false;
});
// Swap
registerCommand(["swap", "sw"], {
    "validator1": {
        0: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid node id '${token.text}'`) : true;
        },
        1: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid connection id '${token.text}'`) : true;
        },
        2: (token) => {
            return token.type != TokenType.Number ? fail(`Invalid connection id '${token.text}'`) : true;
        },
    },
}, null, (tokens) => {
    let nodeId = parseInt(tokens[0].text);
    let sourceId = parseInt(tokens[1].text);
    let targetId = parseInt(tokens[2].text);
    let node = project.getActivePatch().getNode(nodeId);
    let portKey = "";
    if (node != null) {
        let connection = node.findConnection(sourceId);
        if (connection == null) {
            return fail(`There is no connection with id '${sourceId}'`);
        }
        portKey = connection.inputPort.key;
    }
    let result = project.historyManager.doAction("swap-connections", {
        nodeId: nodeId,
        portKey: portKey,
        sourceConnectionId: sourceId,
        targetConnectionId: targetId
    });
    if (result == Errors.NodeNotFound) {
        return fail(`There is no node with the id "${nodeId}"`);
    }
    if (result == Errors.InvalidConnectionId) {
        return fail(`Invalid connection id`);
    }
    return false;
});
// Undo
registerCommand(["undo", "u"], {}, null, (_) => {
    project.historyManager.undo();
    return false;
});
// Redo
registerCommand(["redo", "r"], {}, null, (_) => {
    project.historyManager.redo();
    return false;
});
// History
registerCommand(["history", "h"], {}, null, (_) => {
    let historyLen = project.historyManager.entries.length;
    let curIdx = project.historyManager.historyIndex - 1;
    let from = Math.max(curIdx - 4, 0);
    let end = Math.min(from + 8, historyLen - 1);
    let separator = "-".repeat(24);
    if (from > 0) {
        print(`...1-${from}`);
    }
    for (let i = from; i <= end; i++) {
        let entry = project.historyManager.entries[i];
        let pre = " ";
        if (i == curIdx) {
            print(separator);
            pre = ">";
        }
        let actionDesc = "";
        if (entry.key != null) {
            actionDesc = entry.key;
        }
        else {
            actionDesc = project.historyManager.getAction(entry.key).description;
        }
        print(`${pre} (${i + 1}) ` + actionDesc);
        if (i == curIdx) {
            print(separator);
        }
    }
    if (end < historyLen - 1) {
        print(`...${end + 2}-${historyLen}`);
    }
    return false;
});
// Copy
registerCommand(["copy", "cp"], {}, null, (tokens) => {
    if (tokens.length == 0) {
        return fail(`Too few arguments`);
    }
    let nodeIdList = [];
    for (let i = 0; i < tokens.length; i++) {
        let nodeId = parseInt(tokens[i].text);
        let node = project.getActivePatch().getNode(nodeId);
        if (node == null) {
            return fail(`There is no node with id '${nodeId}'`);
        }
        nodeIdList.push(node.id);
    }
    project.historyManager.doAction("copy-nodes", {
        nodeIdList: nodeIdList
    });
    return false;
});
// Paste
registerCommand(["paste", "pa"], {}, null, (tokens) => {
    project.historyManager.doAction("paste-nodes", {});
    return false;
});
// Namespaces
registerCommand(["namespaces", "ns"], {
    "no-params": {},
    "list-mode": {
        0: (token) => {
            return validateComponentKey(token.text);
        },
    },
}, () => {
    return Object.keys(namespaces);
}, (tokens) => {
    let selectedNSKey = "";
    if (tokens.length == 1) {
        selectedNSKey = tokens[0].text;
    }
    else if (tokens.length > 1) {
        return fail(`Expected 1 argument`);
    }
    if (selectedNSKey != "") {
        let ns = namespaces[selectedNSKey];
        if (ns == null) {
            return fail(`There is no samespace with key '${selectedNSKey}'`);
        }
        for (let c of Object.values(ns.components)) {
            print(c.key);
        }
    }
    else {
        let tree = getNamespaceTree();
        const printBranch = (branch) => {
            for (let [_, sub] of Object.entries(branch)) {
                print(sub.namespace.key);
                printBranch(sub.branch);
            }
        };
        printBranch(tree);
    }
    return false;
});
// Help
registerCommand(["help"], {}, null, (_) => {
    for (let command of commands) {
        print(`${command.name.join(", ")}`);
    }
    return false;
});
//# sourceMappingURL=repl.js.map