"use strict";

import repl from "node:repl";
import fs from "fs";

import { commands, registerCommand, executeCommand, findCommand, lastErrorMessage, parseCmdLine, getProject, setProject, print, fail } from "../dist/repl.js";
import { exportProject, exportPatch } from "../dist/export.js";
import { importProject } from "../dist/import.js";
import Sequencer from "../dist/compiler/sequencer.js";

import "../dist/core_components.js";

registerCommand(
    ["run"],
    {},
    null,
    (_, __) => {
        let sequencer = new Sequencer(getProject(), exportPatch(getProject().getActivePatch(), { exportOutputs: true }));
        let result = sequencer.track();

        const createNodeListStr = (nodes) => {
            let str = "";
            if (nodes != null) {
                for (let node of nodes) {
                    str += (str != "" ? ", " : "") + node.id;
                }
            }
            return str;
        }
        print("--[Result]------------------------------------")
        print("Start nodes: [" + createNodeListStr(result.startNodes) + "]")
        print("Main sequence: [" + createNodeListStr(result.mainSequence) + "]")
        print("Signal sequence map:")
        for (let [nodeId, portMap] of Object.entries(result.signalSequences)) {
            print("  Node #" + nodeId + ":")
            for (let [portKey, nodes] of Object.entries(portMap)) {
                print("    " + portKey + ": [" + createNodeListStr(nodes) + "]");
            }
        }
        print("----------------------------------------------")
        return false;
    }
);

const confirm = (callback) => {
    print("** All changes to this patch will be lost!")
    replServer.question("Are you sure [Y/yes] [n/no]? ", (answer) => {
        if (answer == "y" || answer == "Y") {
            callback(true);
        }
        else {
            callback(false);
        }
        replServer.displayPrompt();
    });
}


registerCommand(
    ["quit", "exit"],
    {},
    null,
    (_, __) => {
        confirm((confirmed) => {
            if (confirmed) {
                print("Bye!");
                replServer.close();
            }
        });
    }
);

// Reset

registerCommand(
    ["reset", "restart"],
    {},
    null,
    (_, __) => {
        confirm((confirmed) => {
            if (confirmed) {
                getProject().reset();
            }
        });
        return false;
    }
);

// Save

registerCommand(
    ["save", "s"],
    {
        "validator1": {
            0: (token) => {
                return !(/([a-zA-Z0-9\s_\\.\-\(\):])+(\.json)$/i).test(token.text) ? 
                    fail(`Invalid file name '${token.text}'`) : true;
            },
        },
    },
    null,
    (tokens, __) => {
        let fileName = tokens[0].text;
        let projectJSON = exportProject(getProject());

        // Save file
        fs.writeFile(fileName, JSON.stringify(projectJSON), function (err) {
            if (err) {
                lastErrorMessage = err;
                return true;
            }
            print('File saved.');
        });
        return false;
    }
);

// Load

registerCommand(
    ["load", "l"],
    {
        "validator1": {
            0: (token) => {
                return !(/([a-zA-Z0-9\s_\\.\-\(\):])+(\.json)$/i).test(token.text) ? 
                    fail(`Invalid file name '${token.text}'`) : true;
            },
        },
    },
    null,
    (tokens, _) => {
        let fileName = tokens[0].text;
        let data;
        try {
            data = fs.readFileSync(fileName, 'utf8');
        } catch (err) {
            return fail(`${err}`);
        }
        let project = importProject(JSON.parse(data), true);
        setProject(project);
        print(`File '${fileName}' loaded.`);
        return false;
    }
);

// tmp

registerCommand(
    ["tmp", "tmp"],
    {
    },
    null,
    (args, _) => {
        print("current project ID counter: " + getProject().idGenerator.id)
        return false;
    }
);

/*
    CLI
*/

function evaluate(command, context, filename, callback) {
    let failed = executeCommand(command);
    if (failed) {
        console.log("Failed:", lastErrorMessage)
    }
    else {
    }
    replServer.displayPrompt();
}

const replServer = repl.start({
    prompt: '> ',
    eval: evaluate,
    completer: completer
});

function completer(line) {
    const tokens = parseCmdLine(line);
    let returnStr = "";
    let hits = [];

    if (tokens.length <= 1) {
        returnStr = line;
        let completions = commands.map(command => command.name[0]);
        hits = completions.filter((c) => c.startsWith(line));
        hits = hits.length > 0 ? hits : completions;
    }
    else {
        let command = findCommand(tokens[0].text);
        if (command != null) {
            let completions = [];
            if (command.autocomplete != null) {
                if (typeof command.autocomplete === "function") {
                    completions = command.autocomplete(tokens);
                }
                else if (Array.isArray(command.autocomplete)) {
                    completions = command.autocomplete;
                }
                else {
                    completions = [command.autocomplete];
                }
            }
            returnStr = tokens[tokens.length - 1].text;
            hits = completions.filter((c) => c.startsWith(returnStr));
        }
    }

    return [hits, returnStr];
  }
