import Node, { MetadataMap, KeyValueMap } from "./node.js";
import { registerAction, ActionData } from "./history_manager.js";
import { ConnectionMapSchema, exportGroup, exportGroups, exportInputConnection, exportNode, exportOutputConnection, GroupSchema, InputConnectionSchema, NodeMapSchema, NodeSchema, OutputConnectionSchema } from "./export.js";
import { importConnections, importNode } from "./import.js";
import Module from "./module.js";
import Patch from "./patch.js";
import Group from "./group.js";
import Project from "./project.js";
import { getRelatedConnections, RelatedConnections } from "./related_connections.js";
import { getProject } from "./repl.js";
import { Normalizer } from "./id_generator.js";
import observer from "./observer.js";

export enum Errors {
    Undetermined = -1,
    Success = 0,
    NodeNotFound,
    SecondNodeNotFound,
    PropertyNotFound,
    PortKeyNotFound,
    SecondPortKeyNotFound,
    InvalidComponentKey,
    ModuleNotFound,
    InvalidConnectionId,
    GroupNotFound,
}

export enum Actions {
    CreateNode = "create-node",
    RemoveNode = "remove-node",
    CreateConnection = "create-connection",
    RemoveConnection = "remove-connection",
    SetProperty = "set-property",
    RemoveProperty = "remove-property",
    SetMetadata = "set-metadata",
    RemoveMetadata = "remove-metadata",
    MoveNodesToModule = "move-nodes-to-module",
    SwapConnections = "swap-connections",
    CreateGroup = "create-group",
    RemoveGroup = "remove-group",
    AddNodesToGroup = "add-nodes-to-group",
    RemoveNodesFromGroup = "remove-nodes-from-group",
    CopyNodes = "copy-nodes",
    PasteNodes = "paste-nodes",
}

export const ORIGIN_PATCH_KEY = "_origin_patch_";
export const ORIGIN_PROJECT_KEY = "_origin_project_";

export type CreateNodeActionData = { componentKey: string; properties?: KeyValueMap; metadata?: MetadataMap };
export type RemoveNodeActionData = { nodeId: number; };
export type CreateConnectionActionData = { fromNodeId: number; fromPortKey: string, toNodeId: number; toPortKey: string, toIndex?: number };
export type RemoveConnectionActionData = { fromNodeId: number; fromPortKey: string, connectionId: number };
export type SetPropertyActionData = { nodeId: number, propertyKey: string, value: any };
export type RemovePropertyActionData = { nodeId: number, propertyKey: string };
export type SetMetadataActionData = { nodeId: number, key: string, value: any };
export type RemoveMetadataActionData = { nodeId: number, key: string };
export type MoveNodesToModuleActionData = { moduleId: number, nodeIdList: number[] };
export type SwapConnectionsActionData = { nodeId: number, portKey: string, sourceConnectionId: number, targetConnectionId: number };
export type CreateGroupActionData = { name: string, nodeIdList?: number[] };
export type RemoveGroupActionData = { groupId: number };
export type AddNodesToGroupActionData = { groupId: number, nodeIdList: number[] };
export type RemoveNodesFromGroupActionData = { groupId: number, nodeIdList: number[] };
export type CopyNodesActionData = { nodeIdList: number[] };
export type PasteNodesActionData = {};

let copiedNodeSchemas: NodeSchema[] = [];

export const hasCopiedNodes = () => {return copiedNodeSchemas.length > 0};

registerAction(Actions.CreateNode, {
    description: "Create Node",
    do: (data: ActionData<CreateNodeActionData>, redo: boolean) => {
        let node: Node | Module | null;
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        if (redo) {
            let id = data.get("id");
            node = patch.createNode(id, data.arg.componentKey, data.arg.properties, data.arg.metadata);
        }
        else {
            let id = (<Project>data.get(ORIGIN_PROJECT_KEY)).idGenerator.next();
            node = patch.createNode(id, data.arg.componentKey, data.arg.properties, data.arg.metadata);
            if (node == null) {
                return Errors.InvalidComponentKey;
            }
            data.set("id", id);
        }
        return Errors.Success;
    },
    undo: (data: ActionData<any>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let id = data.get("id");
        patch.removeNode(id);
    }
});

registerAction(Actions.RemoveNode, {
    description: "Remove Node",
    do: (data: ActionData<RemoveNodeActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId);
        if (node == null) {
            return Errors.NodeNotFound;
        }
        data.set("nodeSchema", exportNode(node));
        patch.removeNode(data.arg.nodeId);
        return Errors.Success;
    },
    undo: (data: ActionData<any>) => {
        let nodeSchema = data.get("nodeSchema");
        importNode(nodeSchema, data.get(ORIGIN_PATCH_KEY));
        importConnections(nodeSchema, data.get(ORIGIN_PATCH_KEY));
    }
});

registerAction(Actions.CreateConnection, {
    description: "Create Connection",
    do: (data: ActionData<CreateConnectionActionData>, redo: boolean) => {
        let id: number;
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let fromNode = patch.getNode(data.arg.fromNodeId) as Node;
        let toNode = patch.getNode(data.arg.toNodeId) as Node;
        let fromPortKey = data.arg.fromPortKey;
        let toPortKey = data.arg.toPortKey;
        let toIndex = data.arg.toIndex;
        if (fromNode == null) {
            return Errors.NodeNotFound;
        }
        if (toNode == null) {
            return Errors.SecondNodeNotFound;
        }
        if (!fromNode.isConnectableOutputPort(fromPortKey)) {
            return Errors.PortKeyNotFound;
        }
        if (!toNode.isConnectableInputPort(toPortKey)) {
            return Errors.SecondPortKeyNotFound;
        }
        if (redo) {
            id = data.get("connectionId");
        } else {
            id = (<Project>data.get(ORIGIN_PROJECT_KEY)).idGenerator.next();
            data.set("connectionId", id);
        }
        if (toIndex === undefined) {
            toIndex = toNode.getInputPort(toPortKey).connections.length;
        }
        fromNode.createConnection(id, fromPortKey, toNode, toPortKey, toIndex);
        return Errors.Success;
    },
    undo: (data: ActionData<CreateConnectionActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let id = data.get("connectionId");
        fromNode.removeOutputConnection(data.arg.fromPortKey, id);
    }
});

registerAction(Actions.RemoveConnection, {
    description: "Remove Connection",
    do: (data: ActionData<RemoveConnectionActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let fromNode = patch.getNode(data.arg.fromNodeId);
        let fromPort = fromNode.getOutputPort(data.arg.fromPortKey);
        if (!redo) {
            let connection = fromPort.getConnection(data.arg.connectionId);
            data.set("portConnectionsSchema", exportOutputConnection(connection));
            data.set("inputIndex", fromPort.getConnection(data.arg.connectionId).getInputIndex());
        }
        fromNode.removeOutputConnection(data.arg.fromPortKey, data.arg.connectionId);
        return Errors.Success;
    },
    undo: (data: ActionData<RemoveConnectionActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let fromNode = patch.getNode(data.arg.fromNodeId) as Node;
        let schema = <OutputConnectionSchema>data.get("portConnectionsSchema");
        let toNode = patch.getNode(schema.toNodeId) as Node;
        let inputIndex = data.get("inputIndex") as number;
        fromNode.createConnection(schema.id, data.arg.fromPortKey, toNode, schema.toPortKey, inputIndex);
    }
});

registerAction(Actions.SetProperty, {
    description: "Set Property",
    do: (data: ActionData<SetPropertyActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let propertyKey = data.arg.propertyKey;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!node.component.hasProperty(propertyKey)) {
            return Errors.PropertyNotFound;
        }
        let oldValue = node.isPropertySet(propertyKey) ?  node.getProperty(propertyKey) : undefined;
        data.set("oldValue", oldValue);
        node.setProperty(propertyKey, data.arg.value);
        return Errors.Success;
    },
    undo: (data: ActionData<SetPropertyActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        node.removeProperty(data.arg.propertyKey);
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setProperty(data.arg.propertyKey, oldValue);
        }
        return Errors.Success;
    }
});

registerAction(Actions.RemoveProperty, {
    description: "Remove Property",
    do: (data: ActionData<RemovePropertyActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let propertyKey = data.arg.propertyKey;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!node.component.hasProperty(propertyKey)) {
            return Errors.PropertyNotFound;
        }
        if (!redo) {
            let oldValue = node.isPropertySet(propertyKey) ? node.getProperty(propertyKey) : undefined;
            data.set("oldValue", oldValue);
        }
        node.removeProperty(data.arg.propertyKey);
        return Errors.Success;
    },
    undo: (data: ActionData<RemovePropertyActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setProperty(data.arg.propertyKey, oldValue);
        }
        return Errors.Success;
    }
});

registerAction(Actions.SetMetadata, {
    description: "Set Metadata",
    do: (data: ActionData<SetMetadataActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let key = data.arg.key;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        let oldValue = node.isMetadataSet(key) ? node.getMetadata(key) : undefined;
        data.set("oldValue", oldValue);
        node.setMetadata(key, data.arg.value);
        return Errors.Success;
    },
    undo: (data: ActionData<SetMetadataActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        //node.removeMetadata(data.arg.key);
        let oldValue = data.get("oldValue");
        //if (oldValue != undefined) {
            node.setMetadata(data.arg.key, oldValue);
        //}
        return Errors.Success;
    }
});

registerAction(Actions.RemoveMetadata, {
    description: "Remove Metadata",
    do: (data: ActionData<RemoveMetadataActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let key = data.arg.key;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        if (!redo) {
            let oldValue = node.isMetadataSet(key) ? node.getMetadata(key) : undefined;
            data.set("oldValue", oldValue);
        }
        node.removeMetadata(data.arg.key);
        return Errors.Success;
    },
    undo: (data: ActionData<RemoveMetadataActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let oldValue = data.get("oldValue");
        if (oldValue != undefined) {
            node.setMetadata(data.arg.key, oldValue);
        }
        return Errors.Success;
    }
});
registerAction(Actions.MoveNodesToModule, {
    description: "Move Nodes To Module",
    do: (data: ActionData<MoveNodesToModuleActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let nodes: Node[] =[];
        let moduleNode = patch.getNode(data.arg.moduleId) as Module;
        if (moduleNode == null) {
            return Errors.ModuleNotFound; 
        }
        // Collect node references
        for (let id of data.arg.nodeIdList) {
            let node = patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            nodes.push(node);
        }
        let externalConnections = getRelatedConnections(nodes, "external");
        // Remove external connections
        for (let [nodeId, connection] of Object.entries(externalConnections)) {
            for (let conn of connection.inputs) {
                patch.getNode(+nodeId).removeInputConnection(conn.toPortKey, conn.id);
            }
            for (let conn of connection.outputs) {
                patch.getNode(+nodeId).removeOutputConnection(conn.fromPortKey, conn.id);
            }
        }
        data.set("externalConnections", externalConnections);
        // Move nodes to module patch
        for (let node of nodes) {
            moduleNode.patch.nodes[node.id] = node;
            //delete patch.nodes[node.id];
            patch.removeNode(node.id, false);
        }
        return Errors.Success;
    },
    undo: (data: ActionData<MoveNodesToModuleActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let moduleNode = patch.getNode(data.arg.moduleId) as Module;
        let nodes: Node[] =[];
        // Collect node references
        for (let id of data.arg.nodeIdList) {
            let node = moduleNode.patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            nodes.push(node);
        }
        // Move nodes from module back to patch
        for (let node of nodes) {
            patch.nodes[node.id] = node;
            delete moduleNode.patch.nodes[node.id];
        }
        // Restore connections
        let externalConnections = data.get("externalConnections") as RelatedConnections;
        for (let [nodeId, connection] of Object.entries(externalConnections)) {
            for (let conn of connection.inputs) {
                let fromNode = patch.getNode(conn.fromNodeId) as Node;
                let toNode = patch.getNode(conn.toNodeId) as Node;
                fromNode.createConnection(conn.id, conn.fromPortKey, toNode, conn.toPortKey);
            }
            for (let conn of connection.outputs) {
                let fromNode = patch.getNode(conn.fromNodeId) as Node;
                let toNode = patch.getNode(conn.toNodeId) as Node;
                fromNode.createConnection(conn.id, conn.fromPortKey, toNode, conn.toPortKey);
            }
        }
        return Errors.Success;
    }
});

registerAction(Actions.SwapConnections, {
    description: "Swap Connections",
    do: (data: ActionData<SwapConnectionsActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        if (node == null) {
            return Errors.NodeNotFound;
        }
        let portKey = data.arg.portKey;
        let index1, index2;
        if (redo) {
            [index1, index2] = data.get("indeces");
        }
        else {
            let port = node.getInputPort(portKey);
            if (port == null || !node.component.hasInputPort(portKey)) {
                return Errors.PortKeyNotFound;
            }
            index1 = port.getConnectionIndex(data.arg.sourceConnectionId);
            index2 = port.getConnectionIndex(data.arg.targetConnectionId);
            if (index1 < 0 || index2 < 0) {
                return Errors.InvalidConnectionId;
            }
        }
        data.set("indeces", [index1, index2]);
        node.swapInputConnections(data.arg.portKey, index1, index2);
        return Errors.Success;
    },
    undo: (data: ActionData<SwapConnectionsActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let node = patch.getNode(data.arg.nodeId) as Node;
        let [index1, index2] = data.get("indeces");
        node.swapInputConnections(data.arg.portKey, index1, index2);
    }
});

registerAction(Actions.CreateGroup, {
    description: "Create Group",
    do: (data: ActionData<CreateGroupActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let id: number;
        if (redo) {
            id = data.get("id") as number;
        }
        else {
            id = getProject().idGenerator.next();
        }
        patch.createGroup(id, data.arg.name || "New Group", "", data.arg.nodeIdList);
        data.set("id", id);
        return Errors.Success;
    },
    undo: (data: ActionData<CreateGroupActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let id = data.get("id") as number;
        patch.removeGroup(id);
    }
});

registerAction(Actions.RemoveGroup, {
    description: "Remove Group",
    do: (data: ActionData<RemoveGroupActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let id = data.arg.groupId;
        let group = patch.groups[id];
        if (group == null) {
            return Errors.GroupNotFound;
        }
        data.set("groupSchema", exportGroup(group));
        data.set("id", group.id);
        patch.removeGroup(id);
        return Errors.Success;
    },
    undo: (data: ActionData<RemoveGroupActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let groupSchema = data.get("groupSchema") as GroupSchema;
        let id = data.get("id") as number;
        patch.createGroup(id, groupSchema.name, groupSchema.comment, groupSchema.nodeIdList);
    }
});

registerAction(Actions.AddNodesToGroup, {
    description: "Add Nodes to Group",
    do: (data: ActionData<AddNodesToGroupActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let group = patch.groups[data.arg.groupId];
        if (group == null) {
            return Errors.GroupNotFound;
        }
        for (let nodeId of data.arg.nodeIdList) {
            group.addNodeId(nodeId);
        }
        return Errors.Success;
    },
    undo: (data: ActionData<AddNodesToGroupActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.removeNodeId(nodeId);
        }
    }
});

registerAction(Actions.RemoveNodesFromGroup, {
    description: "Remove Nodes From Group",
    do: (data: ActionData<RemoveNodesFromGroupActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.removeNodeId(nodeId);
        }
        return Errors.Success;
    },
    undo: (data: ActionData<RemoveNodesFromGroupActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let group = patch.groups[data.arg.groupId];
        for (let nodeId of data.arg.nodeIdList) {
            group.addNodeId(nodeId);
        }
    }
});

registerAction(Actions.CopyNodes, {
    description: "Copy node(s)",
    do: (data: ActionData<CopyNodesActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let nodeIdList = data.arg.nodeIdList
        let nodeSchemas: NodeSchema[] = [];
        for (let id of nodeIdList) {
            let node = patch.getNode(id);
            if (node == null) {
                return Errors.NodeNotFound;
            }
            let nodeSchema = exportNode(node, { excludeInputs: true, excludeOutputs: true });
            for (let dir of ["in", "out"]) {
                let ports = dir == "in" ? node.inputs : node.outputs;
                let connectionMap: ConnectionMapSchema = {};
                for (let [portKey, port] of Object.entries(ports)) {
                    let connections: (InputConnectionSchema | OutputConnectionSchema)[] = [];
                    for (let [_, connection] of port.connections) {
                        let id = dir == "in"
                            ? connection.outputPort.ownerNode.id
                            : connection.inputPort.ownerNode.id;
                        if (nodeIdList.indexOf(id) >= 0) {
                            connections.push(dir == "in" ? exportInputConnection(connection) : exportInputConnection(connection));
                        }
                    }
                    connectionMap[portKey] = connections;
                }
                if (dir == "in") {
                    nodeSchema.inputs = connectionMap;
                }
                else {
                    nodeSchema.outputs = connectionMap;
                }
            }
            nodeSchemas.push(nodeSchema);
        }
        copiedNodeSchemas = nodeSchemas;
        return Errors.Undetermined; // Not undo/redo-able
    },
    undo: (_) => {
    }
});

registerAction(Actions.PasteNodes, {
    description: "Paste nodes(s)",
    do: (data: ActionData<PasteNodesActionData>, redo: boolean) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        let nodeSchemas: NodeSchema[];
        let normalizer = new Normalizer(getProject().idGenerator);
        if (redo) {
            nodeSchemas = data.get("copiedNodeSchemas");
        } else {
            nodeSchemas = copiedNodeSchemas;
            data.set("copiedNodeSchemas", copiedNodeSchemas);
        }
        let createdNodeIdList: number[] = [];
        for (let nodeSchema of copiedNodeSchemas) {
            let node = importNode(nodeSchema, patch, normalizer);
            createdNodeIdList.push(node!.id);
        }
        data.set("createdNodeIdList", createdNodeIdList);
        for (let nodeSchema of nodeSchemas) {
            importConnections(nodeSchema, patch, normalizer);
        }
        return Errors.Success;
    },
    undo: (data: ActionData<PasteNodesActionData>) => {
        let patch = data.get(ORIGIN_PATCH_KEY) as Patch;
        for (let id of data.consume("createdNodeIdList")) {
            patch.removeNode(id);
        }
    }
});
