import { getComponent } from "../component_manager.js";
import { PatchSchema, NodeMapSchema, NodeSchema, OutputConnectionSchema } from "../export.js";
import { NodeType } from "../node.js";
import Project from "../project.js";
import { setProperty } from "../utils.js";

const INTERNAL_SIGNALGATE_COMPONENT_KEY = "internal.signal.gate";
const INTERNAL_SIGNALGATE_PORT_KEY = "signal";

interface SignalSequenceMap {
    [nodeId: number]: {
        [portKey: string]: NodeSchema[];
    }
}

interface SignalSource {
    node: NodeSchema;
    portKey: string;
}

type ResultSequences = {
    startNodes: NodeSchema[];
    mainSequence: NodeSchema[];
    signalSequences: SignalSequenceMap;
}

class Sequencer {
    project: Project;
    patch: PatchSchema;
    pathTrackingMap: {
        [nodeId: number]: {
            counter: number;
            depth: number;
            signalSource: SignalSource | null;
        }
    };

    constructor(project: Project, patch: PatchSchema) {
        this.project = project;
        this.pathTrackingMap = {};
        this.patch = patch;
    }

    getInputCount(node: NodeSchema) {
        let count = 0;
        for (let connections of Object.values(node.inputs || [])) {
            count += connections.length;
        }
        return count;
    }

    getStartNodes(nodes: NodeMapSchema) {
        let startNodes: NodeSchema[] = [];

        for (let node of Object.values(nodes)) {
            let component = getComponent(node.componentKey);
            if (node.type == NodeType.ModuleInput || node.type == NodeType.ModuleOutput) {
                continue;
            }
            // Collect start nodes in modules
            else if (node.type == NodeType.Module) {
                startNodes = [...startNodes, ...this.getStartNodes(node.patch!.nodes)];
                continue;
            }
            else {
                if (component?.data.inputs != null) {
                    continue;
                }
            }
            startNodes.push(node);
        }

        return startNodes;
    }

    track() {
        let __race_count = 0;
        let patchNodes = this.patch.nodes;
        let nodes = this.getStartNodes(this.patch.nodes);
        let depth = 0;
        let result: ResultSequences = {
            startNodes: [...nodes],
            mainSequence: [],
            signalSequences: {},
        };

        while (nodes.length > 0) {
            if (__race_count > 999999) {
                throw("** Infinite loop detected in sequencer track function.")
            }

            let node = nodes.shift() as NodeSchema;
            let component = getComponent(node.componentKey);
            let currentSignalSource = node.id in this.pathTrackingMap ? this.pathTrackingMap[node.id].signalSource : null;

            for (let [portKey, connections] of Object.entries(node.outputs || [])) {
                let signalSource = component?.getOutputPort(portKey)?.signal
                    ? {
                        node: node,
                        portKey: portKey,
                    }
                    : currentSignalSource;

                for (let connection of connections as OutputConnectionSchema[]) {
                    let targetNode = patchNodes[connection.toNodeId];

                    if (!(targetNode.id in this.pathTrackingMap)) {
                        // Init tracking node entry
                        this.pathTrackingMap[targetNode.id] = {
                            counter: this.getInputCount(targetNode),
                            depth: -1,
                            signalSource: signalSource
                        }
                    }
                    else {
                        if (signalSource) {
                            if (this.pathTrackingMap[targetNode.id].signalSource) {
                                if (signalSource.node.id != this.pathTrackingMap[targetNode.id].signalSource!.node.id
                                    || signalSource.portKey != this.pathTrackingMap[targetNode.id].signalSource!.portKey) {

                                    // Merge signals, insert internal signal gate component

                                    let gateNode: NodeSchema = {
                                        "componentKey": INTERNAL_SIGNALGATE_COMPONENT_KEY,
                                        "name": "",
                                        "id": this.project.idGenerator.next(),
                                        "type": NodeType.Internal,
                                    };

                                    setProperty(result.signalSequences, [targetNode.id, portKey, []], gateNode);
                                    //result.signalSequences[targetNode.id][portKey].push(gateNode);
                                    setProperty(result.signalSequences, [signalSource.node.id, signalSource.portKey, []], gateNode);
                                    //result.signalSequences[signalSource.node.id][signalSource.portKey].push(gateNode);

                                    signalSource = this.pathTrackingMap[targetNode.id].signalSource = {
                                        node: gateNode,
                                        portKey: INTERNAL_SIGNALGATE_PORT_KEY
                                    };
                                }
                            }
                            else {
                                this.pathTrackingMap[targetNode.id].signalSource = signalSource;
                            }
                        }
                    }

                    //
                    if (this.pathTrackingMap[targetNode.id].counter > 0) {
                        this.pathTrackingMap[targetNode.id].counter--;
                    }

                    //
                    if (this.pathTrackingMap[targetNode.id].counter == 0) {

                        nodes.push(targetNode);

                        if (signalSource == null) {
                            result.mainSequence.push(targetNode);
                        }
                        else {
                            setProperty(result.signalSequences, [signalSource.node.id, signalSource.portKey, []], targetNode)
                        }
                    }
                }
            }
        }
        return result;
    }
}

export default Sequencer;