import Component from "./component.js";
import Node from "./node.js";

/*

Component identifier key description.
-----------------------------------------

                   core.math.advanced.log
componentKey:                         |-|
keyPart:           |--|
namespaceKey:      |----------------|
componentIdentKey: |--------------------|

*/

export interface Port {
    name: string;
    type: string;
}

export interface Input extends Port {
    isMulti?: boolean;
    required?: boolean;
}

export interface InputMap {
    [portKey: string]: Input
}

export interface Output extends Port {
    signal?: boolean;
}

export interface OutputMap {
    [portKey: string]: Output
}

export interface Property {
    name: string;
    type: string;
    defaultValue: any;
}

export interface Properties {
    [key: string]: Property;
}

export enum Appearance {
    Normal,
    Simple
}

type EvaluatorCallback = (node: Node) => string;

export interface ComponentData {
    name: string;
    icon?: string;
    appearance?: Appearance;
    description?: string;
    inputs?: InputMap;
    outputs?: OutputMap;
    properties?: Properties;
    evaluator?: string | EvaluatorCallback;
    signalEvaluator?: string | EvaluatorCallback;
    onEvaluate?: {
        [name: string]: (node: Node) => string | undefined;
    }
}

interface ComponentsMap {
    [key: string]: Component;
}

export interface NamespaceData {
    name: string;
    defaultIcon?: string;
    description: string;
}

interface NamespaceMap {
    [key: string]: Namespace;
}

export interface NamespaceTree {
    [key: string]: {
        branch: NamespaceTree;
        namespace: Namespace;
    }
}

export let components: ComponentsMap = {};
export let namespaces: NamespaceMap = {};

// export const getPortType = (dir: "input" | "output", componentData: ComponentData, port: string) => {
//     return dir == "input" ?
//         port == flowInput ? flowInput : componentData.inputs?.[port]?.type :
//         port == flowOutput ? flowOutput : componentData.outputs?.[port]?.type;

// }

export const splitIdentKey = (key: string) => {
    let index = key.lastIndexOf(".");
    let nsKey = key.slice(0, index);
    let comKey = key.slice(index + 1);
    return [nsKey, comKey];
}

// export const getComponentFullName = (key: string) => {
//     let [nsKey, comKey] = splitIdentKey(key);
//     let namespace = namespaces[nsKey];
//     return namespace.data.name + "." + namespace.getComponent(comKey).data.name;
// }

export const getIdentKey = (nsKey: string, componentKey: string) => {
    return nsKey + "." + componentKey;
}

export class Namespace {
    key: string;
    components: ComponentsMap = {};
    data: NamespaceData;

    constructor(key: string, data: NamespaceData) {
        this.key = key;
        this.data = data;
    }

    setRequiredAsDefault(componentData: ComponentData) {
        if (componentData.inputs) {
            for (let input of Object.values(componentData.inputs)) {
                if (input.required == null) {
                    input.required = true;
                }
            }
        }
    }

    registerComponent(key: string, componentData: ComponentData) {
        this.setRequiredAsDefault(componentData);
        this.components[key] = new Component(this, key, componentData);
    }

    getComponent(key: string) {
        let data = this.components[key];
        return data;
    }
}

export const getComponent = (key: string) => {
    let [nsKey, comKey] = splitIdentKey(key);
    let namespace = namespaces[nsKey];
    return namespace.getComponent(comKey);
}

export const registerNamespace = (key: string, data: NamespaceData) => {
    let namespace = new Namespace(key, data);
    namespaces[key] = namespace;
    return namespace;
}

export const getNamespaceTree = (): NamespaceTree => {
    let tree: NamespaceTree = {};
    let keys = Object.keys(namespaces).sort();
    for (let key of keys) {
        let namespace = namespaces[key];
        let keyPath = key.split(".");
        let t = tree;
        for (let key of keyPath) {
            if (!(key in t)) {
                t[key] = {
                    branch: {},
                    namespace: namespace
                };
            }
            t = t[key].branch;
        }
    }
    return tree;
}

