import Connection from "./connection.js";
import Group from "./group.js";
import Module from "./module.js";
import Node, { NodeType, PortMap, KeyValueMap } from "./node.js";
import Patch, { GroupMap, NodeMap } from "./patch.js";
import Port from "./port.js";
import Project from "./project.js";

export type ConnectionMapSchema = {[key: string]: (InputConnectionSchema | OutputConnectionSchema)[]};
export type PropertyMapSchema = {[key: string]: PropertySchema};
export type NodeMapSchema = {[key: number]: NodeSchema};
export type ConnectionSchema = InputConnectionSchema & OutputConnectionSchema;

export interface InputConnectionSchema {
    id: number;
    fromNodeId: number;
    fromPortKey: string;
}

export interface OutputConnectionSchema {
    id: number;
    toNodeId: number;
    toPortKey: string;
}

export interface PropertySchema {
    value: any;
}

export interface PropertySchemaMap {
    [key: string]: any;
}

export interface NodeSchema {
    type?: number;
    name?: string;
    id: number;
    componentKey: string;
    inputs?: ConnectionMapSchema;
    outputs?: ConnectionMapSchema;
    properties?: PropertySchemaMap;
    metadata?: PropertySchemaMap;
    patch?: PatchSchema;
}

export interface PatchSchema {
    nodes: NodeMapSchema;
    groups?: GroupSchema[];
}

export interface ProjectSchema {
    name: string;
    patch: PatchSchema;
}

export interface GroupSchema {
    name: string;
    comment: string;
    nodeIdList: number[];
}

// export interface GroupMapSchema {
//     [id: number]: GroupSchema;
// }

export interface ExportOptions {
    excludeInputs?: boolean;
    excludeOutputs?: boolean;
}

export const exportProject = (project: Project, options?: ExportOptions): ProjectSchema => {
    return {
        name: project.name,
        patch: exportPatch(project.mainPatch, {...options, ...{excludeOutputs: true}})
        };
}

export const exportPatch = (patch: Patch, options?: ExportOptions): PatchSchema => {
    let patchSchema: PatchSchema = {
        nodes: exportNodes(patch.nodes, options),
    };
    if (patch.groups) {
        patchSchema.groups = exportGroups(patch.groups);
    }
    return patchSchema;
}

export const exportGroup = (group: Group) => {
    return {
        name: group.name,
        comment: group.comment,
        nodeIdList: [...group.nodeIdList]
    };
}

export const exportGroups = (groupsMap: GroupMap) => {
    let groupSchemas: GroupSchema[] = [];
    for (let group of Object.values(groupsMap)) {
        groupSchemas.push(exportGroup(group));
    }
    return groupSchemas;
}

export const exportNodes = (nodes: NodeMap, options?: ExportOptions): NodeMapSchema => {
    let result: NodeMapSchema = {};
    for (let [id, node] of Object.entries(nodes)) {
        result[+id] = exportNode(node, options);
    }
    return result;
}

export const exportNode = (node: Node, options?: ExportOptions): NodeSchema => {
    let nodeSchema: NodeSchema = {
        id: node.id,
        componentKey: node.componentIdentKey,
    };
    if (node.type != NodeType.Normal) {
        nodeSchema.type = node.type;
    }
    if (node.name && node.name != "") {
        nodeSchema.name = node.name;
    }
    if (node.type == NodeType.Module) {
        nodeSchema.patch = exportPatch((<Module>node).patch, options);
    }
    if (node.properties != null && Object.keys(node.properties).length > 0) {
        nodeSchema.properties = exportKeyValues(node.properties);
    }
    if (node.metadata != null && Object.keys(node.metadata).length > 0) {
        nodeSchema.metadata = exportKeyValues(node.metadata);
    }
    if (options?.excludeInputs !== true && node.inputs != null && Object.keys(node.inputs).length > 0) {
        nodeSchema.inputs = exportConnections(node.inputs, exportInputConnection);
    }
    if (options?.excludeOutputs !== true && node.outputs != null && Object.keys(node.outputs).length > 0) {
        nodeSchema.outputs = exportConnections(node.outputs, exportOutputConnection);
    }
    return nodeSchema;
}

export const exportConnections = (ports: PortMap, connectionExportFn: (connection: Connection) => any): ConnectionMapSchema => {
    let connectionMap: ConnectionMapSchema = {};
    for (let [portKey, port] of Object.entries(ports)) {
        connectionMap[portKey] = exportPortConnections(port, connectionExportFn);
    }
    return connectionMap;
}

export const exportPortConnections = (port: Port, connectionExportFn: (connection: Connection) => any) => {
    let connections: (InputConnectionSchema | OutputConnectionSchema)[] = [];
    for (let [_, connection] of port.connections) {
        connections.push(connectionExportFn(connection));
    }
    return connections;
}

export const exportInputConnection = (connection: Connection) => {
    return {
        id: connection.id,
        fromNodeId: connection.outputPort.ownerNode.id,
        fromPortKey: connection.outputPort.key
    };
}

export const exportOutputConnection = (connection: Connection) => {
    return {
        id: connection.id,
        toNodeId: connection.inputPort.ownerNode.id,
        toPortKey: connection.inputPort.key
    };
}

export const exportConnection = (connection: Connection): ConnectionSchema => {
    let inputPort = connection.inputPort;
    let outputPort = connection.outputPort;
    return {
        id: connection.id,
        fromNodeId: outputPort.ownerNode.id,
        fromPortKey: outputPort.key,
        toNodeId: inputPort.ownerNode.id,
        toPortKey: inputPort.key 
    };
}

export const exportKeyValues = (properties: KeyValueMap) => {
    let result: PropertyMapSchema = {};
    for (let [key, value] of Object.entries(properties)) {
        result[key] = value;
    }
    return result;
}
