import { Errors } from "./actions.js";
import { Events } from "./events.js";
import EventHandler from "./event_handler.js";
import observer from "./observer.js";

const MAX_HISTORY_LENGTH = 50;

export interface HistoryEntry<T = unknown> {
    key: string;
    data: T;
}

export interface HistoryBatchEntry<T = unknown> {
    key: string;
    description: string;
    entries: HistoryEntry<T>[];
}

interface Action<T = unknown> {
    description: string;
    do(data: T, isRedo?: boolean): any;
    undo(data: T): any;
}

export class ActionData<T = unknown> {
    arg: T;
    data: any = {};

    constructor(arg: any) {
        this.arg = arg;
    }

    set(name: string, value: any) {
        this.data[name] = value;
    }
    
    get(name: string) {
        return this.data[name];
    }

    consume(name: string) {
        let value = this.get(name);
        delete this.data[name];
        return value;
    }
}

const actions: { [key: string]: Action<any> } = {};

export const registerAction = (key: string, action: Action) => {
    actions[key] = action;
}

export class HistoryManager extends EventHandler {
    entries: HistoryEntry[] | HistoryBatchEntry[] = [];
    historyIndex: number = 0;
    batch: HistoryEntry[] = [];

    redo() {
        let result = false;
        if (this.canRedo()) {
            let entry = this.entries[this.historyIndex++];

            if ((<HistoryBatchEntry>entry).entries) {
                // Action batch.
                for (let batchEntry of (<HistoryBatchEntry>entry).entries) {
                    let action = actions[batchEntry.key];
                    this.triggerEvent("onBeforeDo", batchEntry.data);
                    result = action.do(batchEntry.data, true);
                    observer.triggerEvent(Events.OnDoAction, (<HistoryEntry>entry).data);
                }
                observer.triggerEvent(Events.OnFinishBatchAction, entry);
            }
            else {
                let action = actions[(<HistoryEntry>entry).key];
                this.triggerEvent("onBeforeDo", (<HistoryEntry>entry).data);
                result = action.do((<HistoryEntry>entry).data, true);
                observer.triggerEvent(Events.OnDoAction, (<HistoryEntry>entry).data);
            }

            observer.triggerEvent(Events.OnUpdateData);
        }
        return result;
    }

    undo() {
        let result = false;
        if (this.canUndo()) {
            let history = this.entries[--this.historyIndex];
            if ((<HistoryBatchEntry>history).entries) {
                let batchHistory = (<HistoryBatchEntry>history).entries;
                for (let i = batchHistory.length - 1; i >= 0; i--) {
                    let entry = batchHistory[i];
                    let action = actions[entry.key];
                    this.triggerEvent("onBeforeUndo", entry.data);
                    result = action.undo(entry.data);
                    observer.triggerEvent(Events.OnUndoAction, (<HistoryEntry>history).data);
                }
                observer.triggerEvent(Events.OnUndoBatchAction, (<HistoryBatchEntry>history));
            }
            else {
                let action = actions[(<HistoryEntry>history).key];
                this.triggerEvent("onBeforeUndo", (<HistoryEntry<any>>history).data);
                result = action.undo((<HistoryEntry>history).data);
                observer.triggerEvent(Events.OnUndoAction, (<HistoryEntry>history).data);
            }
            observer.triggerEvent(Events.OnUpdateData);
        }
        return result;
    }

    pushAction(name: string, data: ActionData) {
        this.entries.length = this.historyIndex++;
        this.entries.push({
            key: name,
            data: data
        } as any);
        if (this.entries.length > MAX_HISTORY_LENGTH) {
            this.entries.shift();
        }
    }

    doAction<T>(name: string, arg?: T, callback?: (data: ActionData<T>) => void) {
        let action = actions[name];
        //let data = new ActionData(arg);
        let data = new ActionData<T>(arg);
        this.triggerEvent("onCreateActionData", data);
        this.triggerEvent("onBeforeDo", data);
        let result = action.do(data);
        if (result == Errors.Success) {
            this.pushAction(name, data);
            observer.triggerEvent(Events.OnDoAction, data);
            observer.triggerEvent(Events.OnUpdateData);
        }
        callback && callback(data);
        return result;
    }

    doBatchAction<T>(key: string, arg?: T, callback?: (data: ActionData<T>) => void) {
        let action = actions[key];
        let data = new ActionData<T>(arg);
        this.triggerEvent("onCreateActionData", data);
        this.triggerEvent("onBeforeDo", data);
        let result = action.do(data);
        this.batch.push({
            key: key,
            data: data
        });
        observer.triggerEvent(Events.OnDoAction, data);
        observer.triggerEvent(Events.OnUpdateData);
        callback && callback(data);
        return result;
    }

    finishBatch(key: string, description: string = "") {
        if (this.batch.length > 0) {
            this.entries.length = this.historyIndex++;
            const entry: HistoryBatchEntry = {
                "description": description,
                key: key,
                entries: [...this.batch]
            };
            (<HistoryBatchEntry[]>this.entries).push(entry);
            observer.triggerEvent(Events.OnFinishBatchAction, entry);
            this.batch.length = 0;
            if (this.entries.length > MAX_HISTORY_LENGTH) {
                this.entries.shift();
            }
            observer.triggerEvent(Events.OnUpdateData);
        }
    }

    resetBatch() {
        this.batch.length = 0;
    }

    cancelBatch() {
        this.batch = [];
    }

    canUndo(): boolean {
        return this.historyIndex > 0;
    }

    canRedo(): boolean {
        return this.entries.length > this.historyIndex;
    }

    reset() {
        this.entries.length = 0;
        this.historyIndex = 0;
        observer.triggerEvent(Events.OnUpdateData);
    }
    
    getAction(name: string) {
        return actions[name];
    }
}
