import observer from "./observer.js";
import Port, { PortType } from "./port.js";
import Connection from "./connection.js";
import { getComponent, Input, Port as ComponentPort } from "./component_manager.js";
import Component, { FLOW_PORT_KEY } from "./component.js";
import Patch from "./patch.js";
import { Events } from "./events.js";

export enum NodeType {
    Normal,
    Module,
    Group,
    ModuleInput,
    ModuleOutput,
    Internal
}

export type PortMap = {
    [portKey: string]: Port;
}

export type KeyValueMap = {
    [key: string]: any;
}

export type MetadataMap = {
    [key: string]: any;
}

export type ConnectionEventData = {
    connection: Connection;
    inputIndex: number | undefined;
}

export type PropertyEventData = {
    node: Node,
    key: string,
    value: any
}

export type MetadataEventData = {
    node: Node,
    key: string,
    value: any
}

export type SwapInputConnectionsEventData = {
    port: Port;
    index1: number;
    index2: number;
}

class Node {
    type = NodeType.Normal;
    name = "";
    id: number;
    componentIdentKey: string;
    component: Component;
    outputs: PortMap = {};
    inputs: PortMap = {};
    properties: KeyValueMap = {};
    metadata: MetadataMap = {};
    ownerPatch: Patch;

    constructor(ownerPatch: Patch, id: number, componentIdentKey: string) {
        this.ownerPatch = ownerPatch;
        this.id = id;
        this.componentIdentKey = componentIdentKey;
        this.component = getComponent(componentIdentKey);
    }

    isConnectableInputPort(key: string) {
        return this.component.hasInputPort(key);
    }

    isConnectableOutputPort(key: string) {
        return this.component.hasOutputPort(key);
    }

    createFlowPort(): ComponentPort {
        return {
            name: "Flow",
            type: "flow"
        };
    }

    getConnectableInputPorts() {
        let flowPort = [FLOW_PORT_KEY, this.createFlowPort()] as [string, Input];
        if (this.component.data.inputs != null) {
            let inputs = Object.entries(this.component.data.inputs);
            inputs.push(flowPort);
            let filtered = inputs.filter(([portKey, portProperties]) => {
                if (portProperties.isMulti) {
                    return true;
                }
                return !((portKey in this.inputs) && this.inputs[portKey].connections.length > 0);
            });
            return filtered.map(([key, data]) => {
                return [key, data.name];
            });
        }
        return [flowPort];
    }

    getConnectableOutputPorts() {
        let flowPort = [FLOW_PORT_KEY, this.createFlowPort()] as [string, Input];
        if (this.component.data.outputs != null) {
            let outputs = Object.entries(this.component.data.outputs);
            outputs.push(flowPort);
            let filtered = outputs.filter(([key, data]) => {
                return !((key in this.outputs) && this.outputs[key].connections.length > 0);
            });
            return filtered.map(([key, data]) => {
                return [key, data.name];
            });
        }
        return [flowPort];
    }

    lazyGetPort(type: PortType, portKey: string) {
        let portMap = type == PortType.Output ? this.outputs : this.inputs;
        if (portMap[portKey] == null) {
            return portMap[portKey] = new Port(portKey, type, this);
        }
        return portMap[portKey];
    }

    getOutputPort(portKey: string) {
        return this.lazyGetPort(PortType.Output, portKey);
    }

    getInputPort(portKey: string) {
        let type = this.component.getInputPort(portKey)?.isMulti ? PortType.InputMulti : PortType.Input;
        return this.lazyGetPort(type, portKey);
    }

    findConnection(connectionId: number) {
        for (let ports of [this.inputs, this.outputs]) {
            for (let port of Object.values(ports)) {
                let connection = port.getConnection(connectionId);
                if (connection != null) {
                    return connection;
                }
            }
        }
        return null;
    }

    removeOldInputConnection(targetNode: Node, targetPortKey: string) {
        let targetPort = targetNode.getInputPort(targetPortKey);
        if (targetPort.type == PortType.Input && targetPort.hasConnections()) {
            targetPort.ownerNode.removeInputConnection(targetPortKey, targetPort.connections[0][0]);
        }
    }

    createConnection(id: number, fromPortKey: string, targetNode: Node, targetPortKey: string, inputIndex?: number) {
        this.removeOldInputConnection(targetNode, targetPortKey);
        let connection = this.getOutputPort(fromPortKey).createConnection(id, targetNode.getInputPort(targetPortKey), inputIndex);
        observer.triggerEvent<ConnectionEventData>(Events.OnCreateConnection, {
            connection: connection,
            inputIndex: inputIndex
        });
        return connection;
    }

    removeOutputConnection(portKey: string, connectionId: number | null) {
        let removedConnections: Connection[] = [];
        if (connectionId == null) {
            for (let [connectionId, connection] of this.outputs[portKey].connections) {
                connection.inputPort.removeConnection(+connectionId);
                this.outputs[portKey].removeConnection(+connectionId);
                removedConnections.push(connection);
                observer.triggerEvent<Connection>(Events.OnRemoveOutputConnection, connection);
            };
        }
        else {
            let connection = this.outputs[portKey].getConnection(connectionId);
            connection.inputPort.removeConnection(connectionId);
            this.outputs[portKey].removeConnection(connectionId);
            removedConnections.push(connection);
            observer.triggerEvent<Connection>(Events.OnRemoveOutputConnection, connection);
        }
        return removedConnections;
    }

    removeInputConnection(portKey: string, connectionId: number | null) {
        if (connectionId == null) {
            for (let [connectionId, connection] of this.inputs[portKey].connections) {
                connection.outputPort.removeConnection(+connectionId);
                this.inputs[portKey].removeConnection(+connectionId)
                observer.triggerEvent<Connection>(Events.OnRemoveInputConnection, connection);
            };
        }
        else {
            let connection = this.inputs[portKey].getConnection(connectionId);
            connection.outputPort.removeConnection(connectionId);
            this.inputs[portKey].removeConnection(connectionId)
            observer.triggerEvent<Connection>(Events.OnRemoveInputConnection, connection);
        }
    }

    swapInputConnections(portKey: string, index1: number, index2: number) {
        let port = this.inputs[portKey];
        let conns = port.connections;
        [conns[index1], conns[index2]] = [conns[index2], conns[index1]];
        observer.triggerEvent<SwapInputConnectionsEventData>(Events.OnSwapInputConnections, {
            port: port,
            index1: index1,
            index2: index2
        });
    }

    isPropertySet(key: string) {
        return key in this.properties;
    }

    setProperty(key: string, value: any) {
        this.properties[key] = value;
        observer.triggerEvent<PropertyEventData>(Events.OnSetProperty, {
            node: this,
            key: key,
            value: value
        });
    }

    getProperty(key: string) {
        let value: any;
        if (this.isPropertySet(key)) {
            value = this.properties[key];
        }
        else {
            let component = getComponent(this.componentIdentKey);
            value = component?.data.properties?.[key]?.defaultValue;
        }
        return value;
    }

    removeProperty(key: string) {
        let value = this.properties[key];
        delete this.properties[key];
        observer.triggerEvent<PropertyEventData>(Events.OnRemoveProperty, {
            node: this,
            key: key,
            value: value
        });
    }

    isMetadataSet(key: string) {
        return key in this.metadata;
    }

    setMetadata(key: string, value: any) {
        if (typeof value === "undefined") {
            this.removeMetadata(key);
        }
        else {
            this.metadata[key] = value;
        }
        observer.triggerEvent<MetadataEventData>(Events.OnSetMetadata, {
            node: this,
            key: key,
            value: value
        });
    }

    getMetadata(key: string) {
        return this.metadata[key];
    }

    removeMetadata(key: string) {
        let value = this.metadata[key];
        delete this.metadata[key];
        observer.triggerEvent<MetadataEventData>(Events.OnRemoveMetadata, {
            node: this,
            key: key,
            value: value
        });
    }

    getName() {
        return this.name || this.component.getName();
    }
}

export default Node;